import unittest
import os
from math import isclose

from .traci_2_1_spreadsheet import Traci21Factors
from ..data import DEFAULT_DATA_PATH

TRACI_2_1 = os.path.join(DEFAULT_DATA_PATH, 'traci_2_1_2014_dec_10_0_test.xlsx')  # note test file

NUM_METHODS = 10


class TraciTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        This is unfortunately very slow (~1.2 s), since xlrd reads the whole workbook ab initio
        :return:
        """
        cls._traci = Traci21Factors(TRACI_2_1)

    def setUp(self):
        self._qi = self._traci.make_interface('quantity')
        self._ii = self._traci.make_interface('index')

    def test_methods(self):
        self.assertEqual(len([m for m in self._ii.lcia_methods()]), NUM_METHODS)

    def test_ecotox(self):
        # note: UUID equality depends on the proper NS UUID being set in the invocation
        tox = self._qi.get('Ecotoxicity, freshwater')
        self.assertEqual(tox.uuid, 'f2bd6435-f825-3e7e-ab40-45ea5d1dc1d5')
        fac = next(self._qi.factors(tox.external_ref, flowable='ethyl carbamate', context='agricultural'))
        self.assertAlmostEqual(fac.value, 1.783221, places=6)

    def test_ipcc_gwp(self):
        T = Traci21Factors(TRACI_2_1)
        gwp = T['Global Warming Air']
        T.add_method_and_compartment(gwp)
        self.assertEqual(len([f for f in T.query.factors(gwp)]), 91)

    def test_compartments(self):
        cs = set(self._qi.compartments(flowable='ddd'))
        self.assertSetEqual(cs, {'agricultural', 'fresh water', 'rural air', 'sea water', 'soil', 'urban air'})


if __name__ == '__main__':
    unittest.main()



"""
The point of the ecoinvent LCIA methods is to ste^H^H^Hadapt the hard work of the Ecoinvent Centre, so as to develop
LCIA methods and results for ecoinvent LCI that match their own results.

2020-01-11
WARNING: this assumes mass as reference quantity by default, which leads to obvious quantity conversion errors for
non-mass-reference indicators like land occupation.

Solution requires access to Ecoinvent metadata in order to determine the reference quantities for flows.
"""

from __future__ import print_function, unicode_literals

from ..archives import BasicArchive
# from lcatools.literate_float import LiterateFloat

from ..entities import LcQuantity

import os
from xls_tools import open_xl
import time

EI_LCIA_VERSION = '3.1'
EI_LCIA_NSUUID = '46802ca5-8b25-398c-af10-2376adaa4623'  # use the same value for all implementations so they lookup

Ecoinvent_Indicators = os.path.join(os.path.dirname(__file__), 'data',
                                    'list_of_methods_and_indicators_ecoinvent_v3.2.xlsx')

EI_LCIA_ERRORS = {
    ('ReCiPe Midpoint (E) (obsolete)', 'water depletion', 'WDP', 
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0}, 
    ('ReCiPe Midpoint (E) w/o LT (obsolete)', 'water depletion w/o LT', 'WDP w/o LT', 
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (H) (obsolete)', 'water depletion', 'WDP', 
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (H) w/o LT (obsolete)', 'water depletion w/o LT', 'WDP w/o LT', 
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (I) (obsolete)', 'water depletion', 'WDP', 
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (E)', 'water depletion', 'WDP',
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (E) w/o LT', 'water depletion w/o LT', 'WDP w/o LT',
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (H)', 'water depletion', 'WDP',
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (H) w/o LT', 'water depletion w/o LT', 'WDP w/o LT',
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
    ('ReCiPe Midpoint (I)', 'water depletion', 'WDP',
     'Water, unspecified natural origin', 'natural resource', 'in water'): {None: 1.0},
}


_EI_LCIA_ERR_FLOWS = set(k[3] for k in EI_LCIA_ERRORS.keys())  # some minimal optimization

class EcoinventLcia(BasicArchive):
    """
    Class to import the Ecoinvent LCIA implementation and construct a flow-cf-quantity catalog.
    The external keys are concatenations of the three
    """
    _ns_uuid_required = True

    _drop_columns = ['Change?']

    _units_sheet = {
        None: 'units',
        '3.8': 'Indicators'
    }
    def _get_units(self):
        try:
            n = self._units_sheet[self._version]
        except KeyError:
            n = self._units_sheet[None]
        b = self._xls.sheet_by_name(n)
        return self._sheet_to_rows(b)

    @staticmethod
    def _sheet_to_rows(sheet):
        g = sheet.get_rows()
        headings = [h.value for h in next(g)]

        rows = []
        for row in g:
            d = dict()
            for i, h in enumerate(headings):
                if h in EcoinventLcia._drop_columns:
                    continue
                d[h.lower()] = row[i].value
            rows.append(d)
        return rows

    def _load_xl_rows(self):
        """
        25+sec just to open_workbook for EI3.1 LCIA (pandas is similar)
        note: this is down to 15.5 sec
        note: back up to 20 sec for EI3.5
        """
        b = self._xls.sheet_by_name(self._sheet_name)

        self._xl_rows = self._sheet_to_rows(b)

    def __init__(self, source, ei_archive=None, ref=None, mass_quantity=None,
                 version=EI_LCIA_VERSION, ns_uuid=EI_LCIA_NSUUID, static=True, **kwargs):
        """
        EI_LCIA_VERSION default is presently 3.1 for the spreadsheet named 'LCIA implementation v3.1 2014_08_13.xlsx'

        :param source:
        :param ei_archive: required to determine reference quantities for flows
        :param ref: hard-coded 'local.ecoinvent.[EI_LCIA_VERSION].lcia'; specify at instantiation to override
        :param mass_quantity:
        :param version: default is '3.1'
        :param ns_uuid: required; default / convention is '46802ca5-8b25-398c-af10-2376adaa4623'
        :param static: this archive type is always static
        :param kwargs: quiet, upstream
        """
        version = str(version)
        if ref is None:
            ref = '.'.join(['local', 'ecoinvent', version, 'lcia'])
        super(EcoinventLcia, self).__init__(source, ref=ref, ns_uuid=ns_uuid, static=True, **kwargs)
        self._xl_rows = []
        self._version = version
        self._wb = None
        self._ei_archive = ei_archive
        if ei_archive is None:
            print('Warning: no ecoinvent archive! Non-mass methods will be broken!')
            mass = mass_quantity or LcQuantity.new('Mass', self._create_unit('kg')[0])
            self.add(mass)
            self._mass = mass
        else:
            self._ei_archive.load_flows()
            self._mass = None

    @property
    def _xls(self):
        if self._wb is None:
            start = time.time()
            self._wb = open_xl(self.source)
            print('Opened workbook; Elapsed time %.3f' % (time.time() - start))
        return self._wb

    @staticmethod
    def _quantity_key(row):
        return ', '.join([row[k] for k in ('method', 'category', 'indicator')])

    @staticmethod
    def _error_key(row):
        return tuple(row[k] for k in ('method', 'category', 'indicator', 'name', 'compartment', 'subcompartment'))

    @property
    def _value_tag(self):
        if float(self._version) >= 3.7:
            return 'cf'
        return 'cf %s' % self._version

    @property
    def _sheet_name(self):
        if self._version == '3.1':
            return 'impact methods'
        else:
            return 'CFs'

    def _create_quantity(self, row):
        """
        here row is a dict from self._xl_rows
        :param row:
        :return:
        """
        key = self._quantity_key(row)
        try_q = self[key]
        if try_q is None:
            unit, _ = self._create_unit(row['unit'])

            q = LcQuantity(key, Name=key, referenceUnit=unit, Comment='Ecoinvent LCIA implementation',
                           Method=row['method'], Category=row['category'], Indicator=row['indicator'])
            self.add(q)
        else:
            q = try_q
        return q

    def _create_all_quantities(self):
        if float(self._version) <= 3.2:
            self._create_all_quantities_3_2()
        else:
            if float(self._version) >= 3.8:
                unit_col = 'unit'
            elif float(self._version) >= 3.7:
                unit_col = 'unitname'
            else:
                unit_col = 'impact score unit'
            qs = self._get_units()
            for row in qs:
                if unit_col != 'unit':
                    row['unit'] = row[unit_col]
                self._create_quantity(row)

    def _create_all_quantities_3_2(self):
        """
        This should not be used-- quantities should be taken from the source file
        :return:
        """
        x = open_xl(Ecoinvent_Indicators)
        w = x.sheet_by_index(0)

        qs = self._sheet_to_rows(w)
        for row in qs:
            self._create_quantity(row)

    def _get_value(self, row):
        if 'Known issue' in row:
            if row['Known issue'] != '':
                return row['Known issue']
        return row[self._value_tag]

    def _check_row(self, row):
        if row['name'] in _EI_LCIA_ERR_FLOWS:
            k = self._error_key(row)
            if k in EI_LCIA_ERRORS:
                err = EI_LCIA_ERRORS[k]
                if self._version in err:
                    return err[self._version]
                elif None in err:
                    return err[None]
        return self._get_value(row)

    def _load_all(self):
        self._create_all_quantities()
        if len(self._xl_rows) == 0:
            self._load_xl_rows()
        for row in self._xl_rows:
            q = self._create_quantity(row)
            try:
                v = self._check_row(row)  # LiterateFloat(self._get_value(row), **row)
            except ValueError:
                print('Skipping row %s' % row)
                continue
            cx = self.tm.add_context((row['compartment'], row['subcompartment']), origin=self.ref)
            fb = row['name']
            if self._ei_archive is not None:
                try:
                    f = next(self._ei_archive.tm.flows_for_flowable(fb))
                    ref_q = f.reference_entity
                    if self[ref_q.external_ref] is None:
                        self.add_entity_and_children(ref_q)
                except StopIteration:
                    print('Unable to find a flow for flowable "%s" -- skipping' % fb)
                    continue
                    # raise KeyError
            else:
                ref_q = self._mass
            self.tm.add_characterization(row['name'], ref_q, q, v, context=cx, origin=self.ref)

        self.check_counter()



import unittest
from .. import LcFlow, LcQuantity
from antelope import CatalogRef


q = LcQuantity.new('Dummy quantity', 'dum', origin='test.origin')
qref = CatalogRef(q.origin, q.external_ref, entity_type='quantity')


class EntityRefTest(unittest.TestCase):
    def test_equality(self):
        self.assertEqual(q, qref)
        self.assertEqual(qref, q)

    def test_hash(self):
        self.assertEqual(hash(q), hash(qref))

    def test_hash_equiv(self):
        f = LcFlow('dummy flow', referenceQuantity=q, origin=q.origin)
        fref = CatalogRef(f.origin, f.external_ref, entity_type='flow')
        s = set()
        s.add(f)
        s.add(fref)
        self.assertEqual(len(s), 1)
        d = dict()
        d[f] = 42
        self.assertIn(fref, d)
        self.assertEqual(d[fref], 42)


if __name__ == '__main__':
    unittest.main()



import unittest
from .. import IPCC_2007_GWP
from ...archives import BasicArchive
from ...exchanges import ExchangeValue
from ...entities.flows import new_flow

from collections import namedtuple

DummyEntity = namedtuple('DummyEntity', ('origin', 'uuid', 'external_ref'))


I = BasicArchive.from_file(IPCC_2007_GWP)
mass = I['Mass']
air = I.tm['air']

def _exch_gen():
    p = DummyEntity('test.null', '1234567', 'dummy process')
    co2 = new_flow('carbon dioxide', mass, origin='test')
    ch4 = new_flow('methane', mass, origin='test')
    yield ExchangeValue(p, co2, 'Output', termination=air, value=34.7)
    yield ExchangeValue(p, ch4, 'Output', termination=air, value=16)


class IpccTestCase(unittest.TestCase):
    def test_lcia(self):
        gwp = I['Global Warming Air']
        res = gwp.do_lcia(_exch_gen())
        self.assertEqual(res.total(), 34.7 + 16*25)


if __name__ == '__main__':
    unittest.main()



"""
This is the oldest code in Antelope (and among my first ever pieces of python code).  Please don't judge.

Written to access ILCD and ecospold archives in either zipped or unzipped or 7z format and GaBi web databases equally

Unfortunately, the 7z format has some kind of bug where time per read grows linearly with the number of reads.
I should report it.
"""

from __future__ import print_function, unicode_literals

import os
import re
import posixpath
import warnings


from zipfile import ZipFile
try:
    from urllib.request import urlopen
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin
    from urllib2 import urlopen

_ext = re.compile('\.([^.]+)$')
protocol = re.compile('^(\w+)://')


def get_ext(fname):
    if bool(_ext.search(fname)):
        return _ext.search(fname).groups()[0].lower()
    else:
        import magic
        typ = magic.from_file(fname)
        if typ.startswith('Zip archive'):
            return 'zip'
        elif typ.startswith('7-zip archive'):
            return '7z'
        else:
            raise ValueError('No filename extension and unsupported file type %s' % typ)


class FileStore(object):
    """
    A strictly local archive of files to be used as a repository
    """
    OK = False

    @staticmethod
    def _create_cache_dir(path):
        d = os.path.join(os.path.dirname(__file__), 'web_cache', protocol.sub('', path))
        if not os.path.isdir(d):
            os.makedirs(d)
        return d

    @staticmethod
    def _access_7z(path):
        try:
            from py7zlib import Archive7z
            fp = open(path, 'rb')
            archive = Archive7z(fp)
        except ImportError:
            warnings.warn('Python package pylzma required to handle 7-zip archives... archive is inactive')
            archive = None
        return archive

    @staticmethod
    def _access_zip(path):
        archive = ZipFile(path)
        return archive

    def __init__(self, path, internal_prefix=None, query_string=None, cache=True):
        """
        Create a FileStore object from a path.  Basically encapsulates the compression algorithm and presents
        a common interface to client code:
        self.listfiles()
        self.countfiles()
        self.readfile(filename)

        writing providers is not presently supported.

        By default remote archives create a local cache to store downloaded files- this cache is mounted as an
        ordinary archive and is checked first when readfile() is called.  You can force a re-download by specifying
        .readfile(..., force_download=True)

        :param path:
        :param internal_prefix: if present, silently absorb / conceal the specified prefix (prepend to requests, trim
         from responses).
        :param query_string: for remote repositories, append the supplied string after '?' in the URL
        :param cache: (True) for remote repositories, cache downloaded files locally and use first
        :return: an archive object
        """

        self.path = path
        self._internal_prefix = internal_prefix

        if bool(protocol.search(path)):
            self.ext = protocol.search(path).groups()[0]
            self.remote = True
            self.compressed = False
            self._archive = None
            self.OK = True
            self._internal_subfolders = []
            print('FileStore refers to a web address using protocol %s' % self.ext)
            self.query_string = query_string
            if self.query_string is not None:
                print(' with query string %s' % self.query_string)
            self.cache = cache
            if self.cache:
                self._cache = FileStore(self._create_cache_dir(self.path), internal_prefix=internal_prefix)
                print(' caching files locally in %s' % self._cache.path)
            return

        self.cache = False
        self.remote = False
        if not os.path.exists(path):
            print('WARNING: path does not resolve.  FileStore will be non-functional.')
            self.compressed = False
            self._archive = None
            self._internal_subfolders = []
            self.ext = None
            return

        if os.path.isdir(path):
            # print('Path points to a directory. Assuming expanded archive')
            if self._internal_prefix is None:
                self.path = os.path.abspath(path) + os.path.sep  # abs reference plus trailing slash
            else:
                self.path = os.path.join(os.path.abspath(path), internal_prefix) + os.path.sep
                self._internal_prefix = None
            self.compressed = False
            self._archive = None
            self.OK = os.access(path, os.R_OK)
            self._internal_subfolders = [x[0][len(self.path):] for x in os.walk(path) if x[0] != path]
        else:
            self.compressed = True
            self.ext = get_ext(path)
            print('Found Extension: %s' % self.ext)
            self._archive = {
                '7z': self._access_7z,
                'zip': self._access_zip
            }[self.ext](path)
            self.OK = self._archive is not False
            if self.OK:
                self._internal_subfolders = {
                    '7z': self._int_dirs_7z,
                    'zip': self._int_dirs_zip
                }[self.ext]()

    @property
    def internal_prefix(self):
        return self._internal_prefix

    @internal_prefix.setter
    def internal_prefix(self, value):
        """
        incrementally add to prefix
        :param value:
        :return:
        """
        if self._internal_prefix is None:
            self._internal_prefix = value
        else:
            self._internal_prefix = self.pathtype.join(self._internal_prefix, value)

    @property
    def pathtype(self):
        if self.compressed or self.remote:
            return posixpath
        else:
            return os.path

    def _int_dirs_7z(self):
        s = set()
        for f in self._archive.files:
            s.add(os.path.split(f.filename)[0])
        return list(s)

    def _int_dirs_zip(self):
        s = set()
        for f in self._archive.namelist():
            s.add(os.path.split(f)[0])
        return list(s)

    def _prefix(self, file):
        if self._internal_prefix is None:
            return file
        return self.pathtype.join(self._internal_prefix, file)

    def _de_prefix(self, file):
        if self._internal_prefix is None:
            return file
        return re.sub('^' + self.pathtype.join(self._internal_prefix, ''), '', file)

    def _gen_files(self):
        """
        Generate files in the archive, removing the internal prefix
        :return:
        """
        if self.remote:
            raise AttributeError('Unable to list files for remote archives')

        if self.compressed:
            if self.ext == '7z':
                if self._internal_prefix is None:
                    lg = (q.filename for q in self._archive.files)
                else:
                    lg = (q.filename for q in self._archive.files if q.filename.startswith(self._internal_prefix))
            elif self.ext == 'zip':
                # filter out directory entries
                if self._internal_prefix is None:
                    lg = (q for q in self._archive.namelist() if q[:-1] not in self._internal_subfolders)
                else:
                    lg = (q for q in self._archive.namelist() if q[:-1] not in self._internal_subfolders
                          and q.startswith(self._internal_prefix))
            else:
                lg = []
        else:
            w = os.walk(self.path)
            lg = []
            for i in w:
                if len(i[2]) > 0:
                    prefix = i[0][len(self.path):]
                    lg.extend([os.path.join(prefix, z) for z in i[2]])
        for l in lg:
            yield self._de_prefix(l)

    def listfiles(self, in_prefix=''):
        """
        generate files in the archive, removing internal prefix, optionally filtering to specified prefix.
        :param in_prefix: optional prefix to limit
        :return:
        """
        if self.remote:
            if self.cache:
                g = (l for l in self._cache.listfiles(in_prefix=in_prefix))
            else:
                g = []
        else:
            g = self._gen_files()

        for l in g:
            if in_prefix is not None:
                if not bool(re.match('^' + in_prefix, l)):
                    continue
            yield l

    def writefile(self, fname, file, mode='wb'):
        if self.remote:
            print('Cannot write remote files.')
            return
        if self.compressed:
            print('Writing to compressed archives not supported.')
            return
        with open(os.path.join(self.path, fname), mode) as fp:
            fp.write(file)

    def readfile(self, fname, force_download=False):
        """
        Have to decide what this does. I think it should return the raw data- since there's no way to get a pointer
        to a file in a generic archive
        :param fname:
        :param force_download: for remote caching archives:
        :return:
        """
        if self.remote:
            if self.cache:
                if force_download:
                    print('Download forced')
                else:
                    try:
                        r = self._cache.readfile(fname)
                        print('Found file in cache: %s' % fname)
                        return r
                    except FileNotFoundError:
                        print('File not found in cache.. downloading')

            url = urljoin(self.path, fname)
            if self.query_string is not None:
                url += '?' + self.query_string
            print('Accessing remote url: %s' % url)
            file = {
                'http': lambda x: urlopen(x),
                None: None
            }[self.ext](url)
            if self.cache:
                self._create_cache_dir(os.path.dirname(url))
                self._cache.writefile(fname, file.read())
                return self._cache.readfile(fname)

            return file.read()

        elif self.compressed:
            try:
                file = {
                    '7z': lambda x: self._archive.getmember(x),
                    'zip': lambda x: self._archive.open(x)
                }[self.ext](repr(self._prefix(fname)).replace("\\\\","/").replace("\\","/")[1:-1]) # Should find better fix
            except KeyError:
                raise FileNotFoundError(self._prefix(fname))
            if file is None:
                return file
            else:
                return file.read()

        else:
            file = open(os.path.join(self.path, self._prefix(fname)), 'rb')
            data = file.read()
            file.close()
            return data



"""
Characterizations are one of the two forms of quantitative information in LCA. A characterization relates two different
quantities of measurement (known as a reference quantity and a query quantity) for one particular flowable substance, in
a specific environmental context and geographic locale.

A QRResult is a namedtuple that documents the result of a "quantity relation" calculation- it reports:
 - the flowable
 - the reference quantity
 - the query quantity
 - the context / compartment
 - the locale
 - the origin of the characterization data
 - the value of the characterization

"""

from collections import namedtuple
from math import isclose

# namedtuple to store the parameters and result of a quantity-relation lookup
# the first 4 are input params, 'locale' is as-found if found, as-specified if not; origin and value are results

QRResult = namedtuple('QRResult', ('flowable', 'ref', 'query', 'context', 'locale', 'origin', 'value'))


class DuplicateCharacterizationError(Exception):
    pass


class MissingReference(Exception):
    """
    A flow must have a reference quantity to be characterized
    """
    pass


class MissingContext(Exception):
    """
    context cannot be None
    """
    pass


class LocaleMismatch(Exception):
    """
    Requested locale is not found
    """
    pass


class Characterization(object):
    """
    A characterization is an affiliation of a flow and a quantity. Characterizations are inherently naively spatialized,
    with factors stored in a dict of locations, and the 'GLO' location being used as the default.

    Note: the !only! place a Characterization is created is in TermManager.add_characterization()
    """

    entity_type = 'characterization'

    @classmethod
    def from_qrresult(cls, qrr):
        return cls(qrr.flowable, qrr.ref, qrr.quantity, qrr.context, qrr.origin, location=qrr.locale, value=qrr.value)

    @classmethod
    def from_flow(cls, flow, quantity, context=None, origin=None, **kwargs):
        if context is None:
            context = flow.context
        if origin is None:
            origin = flow.origin
        return cls(flow['Name'], flow.reference_entity, quantity, context, origin=origin, **kwargs)

    def __init__(self, flow_name, ref_quantity, query_quantity, context, origin=None, **kwargs):
        """

        :param flow_name: a canonical string
        :param quantity:
        :param value: passed to add_value if present
        :param location: 'GLO' passed to add_value if present
        :param origin: of data, if applicable
        :return:
        """
        assert ref_quantity.entity_type == 'quantity', "'ref_quantity' must be an LcQuantity"
        assert query_quantity.entity_type == 'quantity', "'query_quantity' must be an LcQuantity"
        assert ref_quantity.is_lcia_method is False, "'ref_quantity' cannot be an LCIA method"

        self.flowable = str(flow_name)
        self.quantity = query_quantity
        self._ref_q = ref_quantity
        if context is None:
            raise MissingContext('%s, %s, %s' % (flow_name, ref_quantity, query_quantity))
        self._context = context
        self._locations = dict()
        self._origin = origin or query_quantity.origin

        if context.origin is None:
            context.add_origin(self._origin)

        if kwargs:
            self.add_value(**kwargs)
        # self._natural_dirn = None

    @property
    def origin(self):
        return self._origin

    @property
    def context(self):
        return self._context

    @property
    def ref_quantity(self):
        return self._ref_q

    '''
    @property
    def natural_direction(self):
        return self._natural_dirn

    def set_natural_direction(self, c_mgr):
        if self._natural_dirn is not None:
            return
        comp = c_mgr.find_matching(self.flow['Compartment'])
        if comp.is_subcompartment_of(c_mgr.emissions):
            self._natural_dirn = 'Output'
        elif comp.is_subcompartment_of(c_mgr.resources):
            self._natural_dirn = 'Input'
        else:
            self._natural_dirn = False
    '''

    @property
    def is_null(self):
        return len(self._locations) == 0

    @property
    def value(self):
        if 'GLO' in self._locations:
            return self._locations['GLO']
        elif len(self._locations) == 0:
            return None
        elif len(self._locations) == 1:
            return list(self._locations.values())[0]
        else:
            return self._locations

    @property
    def _sval(self):
        if isinstance(self.value, str):
            return '%s' % self.value.replace('\n', '|')
        elif self.value is None:
            return ' --   '
        return '%6.3g' % self.value

    @value.setter
    def value(self, val):
        self._locations['GLO'] = val

    def _lookup(self, locale):
        found_locale = None
        if len(self._locations) > 0:
            if locale in self._locations.keys():
                # we're looking for a locale that exists
                found_locale = locale
            elif 'GLO' in self._locations.keys():
                # we're looking for something specific that doesn't exist, but GLO does exist
                found_locale = 'GLO'
                # today is not the day to write a location best-match finder
        return found_locale

    def query(self, locale):
        found = self._lookup(locale)  # with the addition of LocaleMismatch, we should do away with this
        if found is None:
            raise LocaleMismatch(list(self._locations.keys()))
        return QRResult(self.flowable, self.ref_quantity, self.quantity, self.context, found,
                        self.origin, self._locations[found])

    def __getitem__(self, item):
        if item == 'quantity':  # f%&(@*$ marshmallow hack
            return self.quantity
        if item == 'flow':  # ibid.
            return self.flowable
        found = self._lookup(item)
        if found is None:
            return 0.0
        return self._locations[found]

    def __setitem__(self, key, value):
        if key in self._locations:
            if self._locations[key] == value:
                return  # just skip if they are the same
            elif isclose(self._locations[key], value):
                return
            elif isclose(self._locations[key], value, rel_tol=1e-6):
                print('%e %e' % (self._locations[key], value))
                return
            if isinstance(self._locations[key], str):
                pval = '%s (incoming: %s)' % (self._locations[key], value)
            else:
                pval = '%g (incoming: %s)' % (self._locations[key], value)
            raise DuplicateCharacterizationError('%s: Characterization value already present! %s = %s\n%s' %
                                                 (self.quantity, key, pval, self.flowable))
        self._locations[key] = value

    def update_values(self, **kwargs):
        self._locations.update(kwargs)

    def add_value(self, value=None, location=None, overwrite=False):
        if location is None:
            location = 'GLO'
        if overwrite:
            if location in self._locations:
                self._locations.pop(location)
        self[location] = value

    '''
    def scale(self, factor):
        for k, v in self._locations.items():
            self._locations[k] = v * factor
    '''

    def locations(self):
        return self._locations.keys()

    def list_locations(self):
        return '; '.join([k for k in self.locations()])

    def __hash__(self):
        return hash((self.flowable, self.ref_quantity.external_ref, self.quantity.external_ref, self.context))

    def __eq__(self, other):
        """
        Returns true if all of other's location-specific values equal self's values for the same location
        :param other:
        :return:
        """
        if other is None:
            return False
        if ((self.flowable == other.flowable) &
                (self.quantity == other.quantity)):
            if all(self[l] == other[l] for l in other.locations()):
                return True
        return False

    def __str__(self):
        if self.is_null:
            return '%s has %s %s' % (self.flowable, self.quantity, self.quantity.reference_entity)
        scs = []
        for k, v in self._locations.items():
            if isinstance(v, str):
                scs.append("'%s' [%s]" % (v, k))
            else:
                scs.append('%6.3g [%s]' % (v, k))

        return '%s [%s / %s] %s: %s (%s)' % ('\n'.join(scs),
                                             self.quantity.unit, self.ref_quantity.unit, self.flowable, self.context,
                                             self.quantity.name)

    def __repr__(self):
        if self.is_null:
            return '%s(%s, %s, %s: null (%s))' % (self.__class__.__name__, self.flowable,
                                                 self.ref_quantity.unit, self.context, self.quantity.unit)
        if len(self._locations) > 1:
            val = '%s (+%d)' % (self._sval, len(self._locations) - 1)
        else:
            val = '%s' % self._sval
        return '%s(%s, %s, %s: %s (%s))' % (self.__class__.__name__, self.flowable,
                                           self.ref_quantity.unit,
                                           self.context, val, self.quantity.unit)

    def q_view(self):
        if self.quantity is self.ref_quantity:
            ref = '(*)'
        else:
            ref = ' | '
        return '%25.25s [%s / %s]%s%s' % (self._sval, self.quantity.unit,
                                          self.ref_quantity.unit, ref,
                                          self.quantity.name)

    '''
    def tupleize(self):
        return self.flow.get_uuid(), self.quantity.get_uuid()
    '''

    def serialize(self, values=False, concise=False):
        """
        The "concise" option is used within term manager when query quantity, context, and flowable are already
        serialized
        :param values:
        :param concise:
        :return:
        """
        j = {
            'ref_quantity': self.ref_quantity.external_ref
        }
        if self.ref_quantity is self.quantity:
            j['isReference'] = True
        else:
            if not concise:
                j['entityType'] = self.entity_type
                j['query_quantity'] = self.quantity.external_ref
                j['context'] = str(self.context)
                j['flowable'] = self.flowable
            if values:
                if self.value is not None:
                    j['value'] = self._locations
        return j



from ..autorange import AutoRange

import unittest


values = [-0.34, 12e9, .0000867]
scales = [1000, 1.0e-9, 1.0e6]
adj_values = [-340, 12, 86.7]
units = ['kg CO2eq', 'mol H+', 'MMT']
correct_units = [['g CO2eq', 'Tg CO2eq', 'mg CO2eq'],
                 ['mmol H+', 'Gmol H+', 'umol H+'],
                 ['kMT', 'PMT', 'MT']]


class AutoRangeTestCase(unittest.TestCase):
    def test_conversion(self):
        for i, val in enumerate(values):
            fm = 'Failed on %g [item #%d]' % (val, i)
            a = AutoRange(val)
            self.assertEqual(a.scale, scales[i], fm)
            self.assertEqual(a.adjust(val), adj_values[i], fm)

    def test_units(self):
        for i, val in enumerate(values):
            fm = 'Failed on %g [item #%d]' % (val, i)
            a = AutoRange(val)
            for j, unit in enumerate(units):
                self.assertEqual(a.adj_unit(unit), correct_units[j][i], fm)


if __name__ == '__main__':
    unittest.main()



"""
A repository of typed entities, retrievable by their external reference

Entity object API:
  entity.entity_type --> string used for groupung
  entity.external_ref --> lookup name
  entity.origin --> one-time settable parameter, set by the entity store
  entity.validate() --> must return True [for valid entities and False for invalid ones]
  entity.name --> printable name

Optional:
  entity.uuid --? used for entity retrieval

"""
from __future__ import print_function, unicode_literals

import uuid
import re
import os
from datetime import datetime

from collections import defaultdict

from antelope import local_ref
from ..from_json import to_json


# CatalogRef = namedtuple('CatalogRef', ['archive', 'id'])


ref_regex = re.compile('[a-z0-9_]+(\.[a-z0-9_]+)*', flags=re.IGNORECASE)


uuid_regex = re.compile('([0-9a-f]{8}-?([0-9a-f]{4}-?){3}[0-9a-f]{12})', flags=re.IGNORECASE)


def to_uuid(_in):
    if _in is None:
        return _in
    if isinstance(_in, int):
        return None
    try:
        g = uuid_regex.search(_in)  # using the regexp test is 50% faster than asking the UUID library
    except TypeError:
        if isinstance(_in, uuid.UUID):
            return str(_in)
        g = None
    if g is not None:
        return g.groups()[0]
    '''
    # no regex match- let's see if uuid.UUID can handle the input  
    try:
        _out = uuid.UUID(_in)
    except ValueError:
        return None
    return str(_out)
    '''    ## NOTE: This is costly because it requires to instantiate a UUID for EVERY query, especially those that are
    # already probably not valid UUIDs! There is every reason to expect the input is a string, and our regex already
    # matches even non-RFC-compliant UUID strings. I'm going to leave it out for now
    return None



class SourceAlreadyKnown(Exception):
    pass


class EntityExists(Exception):
    pass


class InvalidSemanticReference(Exception):
    pass


class ReferenceCreationError(Exception):
    pass


class EntityStore(object):
    _entity_types = ()  # must be overridden
    '''
    _ns_uuid_required: specifies whether the archive must be supplied an ns_uuid (generally, archives that are
    expected to generate persistent, deterministic IDs must have an externally specified ns_uuid)
     If False: random ns_uuid generated if none is supplied
     If True: ns_uuid must be supplied as an argument, will raise exception if missing
     If None: ns_uuid forced to None - store does not have ns_uuid capabilities
    '''
    _ns_uuid_required = False
    _origin = None  # can be set when a catalog is assigning a ref

    def _ref_to_uuid(self, key):
        """
        This tries to find a UUID from a ref. Not sure what this is good for.
        by default, to_uuid just returns a string matching the regex, or failing that, tries to generate a string
        using uuid.UUID(key)
        :param key:
        :return:
        """
        u = to_uuid(key)  # check if key is already a uuid
        if u is None:
            return self._ref_to_nsuuid(key)
        return u

    def _ref_to_nsuuid(self, key):
        if self._ns_uuid is None:
            return None
        if isinstance(key, int):
            key = str(key)
        return str(uuid.uuid3(self._ns_uuid, key))

    def _ref_to_key(self, key):
        """
        This method always returns a valid key into _entities, or None.  May be overridden.

        :param key:
        :return:
        """
        if key in self._entities:
            return key
        uu = self._ref_to_uuid(key)
        if uu is not None:
            if uu in self._entities:
                return uu

    def get_uuid(self, key):
        """
        Deprecated.
        :param key:
        :return:
        """
        return self._ref_to_uuid(key)

    def _set_ns_uuid(self, ns_uuid):
        print('%s: Setting NSUUID (%s) %s' % (self.ref, self._ns_uuid_required, ns_uuid))
        if self._ns_uuid_required is None:
            if ns_uuid is not None:
                print('Ignoring ns_uuid specification')
            return None
        else:
            if ns_uuid is None:
                if self._ns_uuid_required is True:
                    raise AttributeError('ns_uuid specification required')
                elif self._ns_uuid_required is False:
                    return uuid.uuid4()
            else:
                if isinstance(ns_uuid, uuid.UUID):
                    return ns_uuid
                return uuid.UUID(ns_uuid)

    def __init__(self, source, ref=None, quiet=True, upstream=None, static=False, dataReference=None, ns_uuid=None,
                 no_validate=None,
                 **kwargs):
        """
        An EntityStore is a provenance structure for a collection of entities.  Ostensibly, an EntityStore has a single
        source from which entities are collected.  The source is a resolvable URI that indicates a data resource from
        which data describing the entities can be extracted.  The exact manner of extracting data from resources is
        subclass-dependent.

        Internally, all entities are stored with UUID keys.  If the external references do not contain UUIDs, it is
        recommended to derive a UUID3 using an archive-specific, stable namespace ID.  The class-level
        _ns_uuid_required attribute governs this option:
         - if True, an ns_uuid argument must be provided when the class is instantiated.  This is consistent with a
           use case in which it is desirable to have predictable, fixed UUIDs (i.e. to interface with a data system
           that requires stable UUIDs)

         - if False, a random ns_uuid is generated, and used to create a UUID anytime an entity is given a non-UUID
           external_ref

         - if None, UUID3 are not used and any supplied ns_uuid argument is ignored. external_refs must always be UUIDs.

        There is still some refactoring to be done, to try to eliminate the need for externally visible UUIDs anywhere.

        An archive has a single semantic reference that describes the data context from which its native entities
        were gathered.  The reference is given using dot-separated hierarchical terms in order of decreasing
        semantic significance from left to right.  The leftmost specifier should describe the maintainer of the
         resource (which defaults to 'local' when a reference argument is not provided), followed by arbitrarily
         more precise specifications. Some examples are:
         local.lcia.traci.2.1.spreadsheet
         ecoinvent.3.2.undefined

        The purpose for the source / reference distinction is that in principle many different sources can all provide
        the same semantic content: for instance, ecoinvent can be accessed from the website or from a file on the
        user's computer.  In principle, if the semantic reference for two archives is the same, the archives should
        contain excerpts of the same data, even if drawn from different sources.

        An entity is uniquely identified by its link property, which is made from concatenating the semantic origin and
        a stable reference known as an 'external_ref', as 'origin/external_ref'.  The first slash is the delimiter
        between origin and reference. Examples:

        elcd.3.2/processes/00043bd2-4563-4d73-8df8-b84b5d8902fc
        uslci.ecospold/Acetic acid, at plant

        Note that the inclusion of embedded whitespace, commas, and other characters indicate that these semantic
        references are not proper URIs.

        It is hoped that the user community will help develop and maintain a consistent and easily interpreted
        namespace for semantic references.  If this is done, it should be possible to identify any published entity
        with a concise reference.

        When an entity is first added to an archive, it is assigned that archive's *reference* as its origin, following
        the expectation that data about the same reference from different sources is the same data.

        When an entity with a different origin is added to an archive, it is good practice to add a mapping from that
        origin to its source in the receiving archive's "catalog_names" dictionary.  However, since the entity itself
        does not know its archive's source, this cannot be done automatically.

        :param source: physical data source-- where the information is being drawn from
        :param ref: optional semantic reference for the data source. gets added to catalog_names.
        :param quiet:
        :param upstream:
        :param static: [False] whether archive is expected to be unchanging.
        :param dataReference: alternative to ref
        :param ns_uuid: required to store entities by common name.  Used to generate uuid3 from string inputs.
        :param no_validate: if True, skip validation on entity add
        :param kwargs: any other information that should be serialized with the archive
        """

        self._source = source
        if ref is None:
            if dataReference is None:
                ref = local_ref(source)
            else:
                ref = dataReference

        self._entities = {}  # uuid-indexed list of known entities

        self._quiet = quiet  # whether to print out a message every time a new entity is added / deleted / modified

        self._serialize_dict = kwargs  # this gets added to

        self._counter = defaultdict(int)
        self._ents_by_type = defaultdict(set)
        self._upstream = None
        self._no_validate = no_validate

        self._loaded = False
        self._static = static
        self._descendant = False

        if upstream is not None:
            self.set_upstream(upstream)

        self._catalog_names = defaultdict(set)  # this is a place to map semantic references to data sources
        self._add_name(ref, source)
        self._serialize_dict['dataReference'] = ref

        self._ns_uuid = self._set_ns_uuid(ns_uuid)
        if self._ns_uuid is not None:
            self._serialize_dict['ns_uuid'] = str(self._ns_uuid)

    def _add_name(self, ref, source, rewrite=False):
        """
        A source is not allowed to provide multiple semantic references
        a ref must match the regexp ([A-Za-z0-9_]+(\.[A-Za-z0-9_])*)
        :param ref:
        :param source:
        :param rewrite: [False] if True, if SourceAlreadyKnown, re-assign the source to the new ref. This may result
        in the archive's ref changing, and should only be used when an authoritative source-ref pair is supplied
        (e.g. a JSON file that was loaded into the archive)
        :return:
        """
        if not ref_regex.match(ref):
            raise InvalidSemanticReference('%s' % ref)
        for k, s in self._catalog_names.items():
            if source in s and source is not None:
                if source == self.source and k == local_ref(self.source):
                    '''if we're trying to add our own source and ref to the name index, and the source is currently
                    registered to the default local_ref, then we override it
                    '''
                    self._catalog_names[ref] = self._catalog_names.pop(k)
                    return
                if k == ref or ref.startswith(k):
                    return
                if rewrite:
                    self._catalog_names[k].remove(source)
                    print('%s: <source removed>' % k)
                else:
                    raise SourceAlreadyKnown('Source %s already registered to name %s (vs: %s)' % (source, k, ref))
        print('%s: %s' % (ref, source))
        self._catalog_names[ref].add(source)
        if ref == self.ref and self.source is None and rewrite:
            self._source = source

    def add_new_source(self, new_ref, new_source):
        self._add_name(new_ref, new_source, rewrite=False)

    @property
    def source(self):
        """
        The catalog's original source is the "master descriptor" of the catalog's content. This is required for
        subclass methods to work properly, in the event that the original source is called upon.
        :return:
        """
        return self._source

    def _set_source(self, new_ref, new_source):
        self._source = new_source
        self._add_name(new_ref, new_source)
        self._descendant = True

    def set_origin(self, origin):
        self._serialize_dict['dataReference'] = origin
        self._add_name(origin, self.source, rewrite=True)
        self._origin = origin

    @property
    def ref(self):
        if self._origin is not None:
            return self._origin
        try:
            return next(k for k, s in self._catalog_names.items() if self.source in s)
        except StopIteration:
            return local_ref(self.source)

    @property
    def catalog_names(self):
        for k in self._catalog_names.keys():
            yield k

    @property
    def names(self):
        """
        Return a mapping of data source to semantic reference, based on the catalog_names property.  This is used by
        a catalog interface to convert entity origins from physical to semantic.

        If a single data source has multiple semantic references, only the most-downstream one will be kept.  If there
        are multiple semantic references for the same data source in the same archive, one will be kept at random.
        This should be avoided and I should probably test for it when setting catalog_names.
        :return:
        """
        if self._upstream is None:
            names = dict()
        else:
            names = self._upstream.names

        for k, s in self._catalog_names.items():
            for v in s:
                names[v] = k
        return names

    def get_sources(self, name):
        s = self._catalog_names[name]
        if len(s) == 0:
            for k, ss in self._catalog_names.items():
                if k.startswith(name):
                    s = s.union(ss)
        for d in s:
            yield d

    def construct_new_ref(self, signifier):
        today = datetime.now().strftime('%Y%m%d')
        if signifier is None:
            new_tail = today
        else:
            if not bool(re.match('[A-Za-z0-9_-]+', signifier)):
                raise ValueError('Invalid signifier %s' % signifier)
            new_tail = '.'.join([signifier, datetime.now().strftime('%Y%m%d')])

        if len(self.ref.split('.')) > 2:  # must be true to be postfixing a postfix
            old_tail = '.'.join(self.ref.split('.')[-2:])
            if old_tail.startswith(new_tail):
                hm = '.' + datetime.now().strftime('-%H%M')
                if old_tail.startswith(new_tail + hm):
                    hm += datetime.now().strftime('%S')
                    if old_tail.startswith(new_tail + hm):
                        raise ReferenceCreationError('HMS? %s', (self.ref, hm))
                new_tail += hm

            elif old_tail.find('.' + today) >= 0 and signifier is not None:
                # don't reprint the date if it already shows up
                new_tail = signifier

        new_ref = '.'.join([self.ref, new_tail])
        return new_ref

    def create_descendant(self, archive_path, signifier=None, force=False):
        """
        Saves the archive to a new source with a new semantic reference.  The new semantic ref is derived by
         (a) first removing any trailing ref that matches [0-9]{8+}
         (b) appending the descendant signifier
         (c) appending the current date in YYYYMMDD format

        After that:
         1. The new semantic ref is added to catalog_names,
         2. the source is set to archive_path/semantic.ref.json.gz,
         3. load_all() is executed,
         4. the archive is saved to the new source.

        :param archive_path: where to store the archive
        :param signifier: A nonzero-length string matching [A-Za-z0-9_-]+.  If not supplied, then the semantic ref is
        unchanged except for the date tag.
        :param force: overwrite if file exists
        :return: new semantic ref.
        """
        if not os.path.exists(archive_path):
            os.makedirs(archive_path)

        new_ref = self.construct_new_ref(signifier)
        if new_ref == self.ref:
            raise KeyError('Refs are the same!')  # KeyError bc it's a key in catalog_names

        new_filename = new_ref + '.json.gz'
        new_source = os.path.join(archive_path, new_filename)
        if os.path.exists(new_source):
            if force:
                print('Overwriting existing archive')
            else:
                raise EnvironmentError('File %s exists: force=True to overwrite' % new_source)

        try:
            self.load_all()
        except NotImplementedError:
            pass
        self._set_source(new_ref, new_source)
        self.write_to_file(new_source, gzip=True, complete=True)
        return new_ref

    @property
    def static(self):
        return self._static or self._loaded

    '''
    @property
    def ref(self):
        """
        Deprecated.  Archives have a source; catalogs have a ref.
        :return:
        """
        return self._source
    '''

    def entities(self):
        for v in self._entities.values():
            yield v

    def set_upstream(self, upstream):
        assert isinstance(upstream, EntityStore)
        if upstream.source != self.source:
            self._serialize_dict['upstreamReference'] = upstream.ref
        self._upstream = upstream

    '''
    def truncate_upstream(self):
        """
        BROKEN! / deprecated
        removes upstream reference and rewrites entity uuids to match current index. note: deprecates the upstream
        upstream_
        :return:
        """
        # TODO: this needs to be fixed: truncate needs localize all upstream entities (retaining their origins)
        for k, e in self._entities.items():
            e._uuid = k
        self._upstream = None
        if 'upstreamReference' in self._serialize_dict:
            self._serialize_dict.pop('upstreamReference')
    '''

    def _print(self, *args):
        if self._quiet is False:
            print(*args)

    def __str__(self):
        count = sum(len(v) for v in self._ents_by_type.values())
        s = '%s with %d entities at %s' % (self.__class__.__name__, count, self.source)
        if self._upstream is not None:
            s += ' [upstream %s]' % self._upstream.__class__.__name__
        return s

    def _get_entity(self, key):
        """
        the fundamental method- retrieve an entity from LOCAL collection by key, nominally a UUID string.

        If the string is not found, raises KeyError.
        :param key: a uuid
        :return: the LcEntity or None
        """
        if key in self._entities:
            return self._entities[key]
        raise KeyError(key)

    def __contains__(self, item):
        return item in self._entities

    def __getitem__(self, item):
        """
        CLient-facing entity retrieval.  item is a key that can be converted to a valid UUID from self._ref_to_key()--
         either a literal UUID, or a string containing something matching a naive UUID regex.

        First checks upstream, then local.

        Returns None if nothing is found

        :param item:
        :return:
        """
        if item is None:
            return None
        if self._upstream is not None:
            e = self._upstream[item]
            if e is not None:
                return e
        try:
            if isinstance(item, int) and self._ns_uuid is not None:
                return self._get_entity(self._ref_to_nsuuid(item))
            return self._get_entity(self._ref_to_key(item))
        except KeyError:
            return None

    def _ensure_valid_refs(self, entity):
        """
        Hook to validate the incoming entity's references-- namely, to set its uuid
        :param entity:
        :return:
        """
        if hasattr(entity, 'uuid') and entity.uuid is None:
            uu = self._ref_to_uuid(entity.external_ref)
            if uu is not None:
                entity.uuid = uu

    def _add(self, entity, key, quiet=False):
        self._ensure_valid_refs(entity)

        if key is None:
            raise ValueError('Key not allowed to be None')
        if key in self._entities:
            raise EntityExists('Entity already exists: %s' % key)

        if entity.entity_type not in self._entity_types:
            raise TypeError('Entity type %s not valid!' % entity.entity_type)

        if entity.is_entity and not self._no_validate:
            if not entity.validate():
                raise ValueError('Entity fails validation: %s' % repr(entity))

        if not (self._quiet or quiet):
            print('Adding %s entity with %s: %s' % (entity.entity_type, key, entity['Name']))
        if entity.origin is None:
            # TODO: uncomment / enforce this
            # assert self._ref_to_key(entity.external_ref) == key, 'entity uuid must match origin repository key!'
            entity.origin = self.ref
        self._entities[key] = entity
        if self._ns_uuid is not None:  # ensure UUID3s work even if custom UUIDs are specified
            nsuuid = self._ref_to_uuid(entity.external_ref)
            if nsuuid is not None and nsuuid not in self._entities:
                self._entities[nsuuid] = entity
        self._counter[entity.entity_type] += 1
        self._ents_by_type[entity.entity_type].add(key)  # it's not ok to change an entity's type

    def check_counter(self, entity_type=None):
        if entity_type is None:
            [self.check_counter(entity_type=k) for k in self._entity_types]
        else:
            print('%d new %s entities added (%d total)' % (self._counter[entity_type], entity_type,
                                                           self.count_by_type(entity_type)))
            self._counter[entity_type] = 0

    def find_partial_id(self, uid, upstream=False, startswith=True):
        """
        :param uid: is a fragmentary (or complete) uuid string.
        :param upstream: [False] whether to look upstream if it exists
        :param startswith: [True] use .startswith instead of full regex
        :return: result set
        """
        if startswith:
            def test(x, y):
                return y.startswith(x)
        else:
            def test(x, y):
                return bool(re.search(x, y))
        result_set = [v for k, v in self._entities.items() if test(uid, k)]
        if upstream and self._upstream is not None:
            result_set += self._upstream.find_partial_id(uid, upstream=upstream, startswith=startswith)
        return result_set

    def _fetch(self, entity, **kwargs):
        """
        Dummy function to fetch from archive. MUST be overridden.
        Can't fetch from upstream.
        :param entity:
        :return:
        """
        raise NotImplementedError

    def retrieve_or_fetch_entity(self, key, **kwargs):
        """
        Client-facing function to retrieve entity by ID, first checking in the archive, then from the source.

        Input is flexible-- could be a UUID or key (partial uuid is just not useful)

        :param key: the identifying string (uuid or external ref)
        :param kwargs: used to pass provider-specific information
        :return:
        """

        if key is not None:
            entity = self.__getitem__(key)  # this checks upstream if it exists
            if entity is not None:
                # retrieve
                return entity

        # fetch
        return self._fetch(key, **kwargs)

    def get(self, key):
        return self.retrieve_or_fetch_entity(key)

    def validate_entity_list(self):
        """
        This whole thing is crufty and untested and never used and should be abandoned
        :return:
        """
        count = 0
        for k, v in self._entities.items():
            valid = True
            '''
            # 1: confirm key is a UUID
            if not isinstance(k, uuid.UUID):
                print('Key %s is not a valid UUID.' % k)
                valid = False
            '''
            if v.origin is None:
                print("%s: No origin!" % k)
                valid = False

            if v.origin == self.source:
                # 2: confirm entity's external key maps to its uuid
                if self._ref_to_uuid(v.external_ref) != v.uuid:
                    print("%s: Key doesn't match UUID in origin!" % v.external_ref)
                    valid = False

            # confirm entity is dict-like with keys() and with a set of common keys
            try:
                valid = valid & v.validate()
            except AttributeError:
                print('Key %s: not a valid LcEntity (no validate() method)' % k)
                valid = False

            if valid:
                count += 1
        print('%d entities validated out of %d' % (count, len(self._entities)))
        return count

    def _load_all(self, **kwargs):
        """
        Must be overridden in subclass
        :return:
        """
        raise NotImplementedError

    def load_all(self, **kwargs):
        if self._loaded is False:
            print('Loading %s' % self.source)
            self._load_all(**kwargs)
            self._loaded = True

    def entities_by_type(self, entity_type):
        for u in sorted(self._ents_by_type[entity_type]):
            yield self._entities[u]

    def count_by_type(self, entity_type):
        return len(self._ents_by_type[entity_type])

    @property
    def init_args(self):
        return self._serialize_dict

    def serialize(self, **kwargs):
        j = {
            'dataSourceType': self.__class__.__name__,
            'dataSource': self.source,
            'catalogNames': {k: sorted(filter(None, s)) for k, s in self._catalog_names.items()},
            'initArgs': self._serialize_dict
        }
        return j

    def _serialize_all(self, **kwargs):
        """
        To be overridden-- specify args necessary to make a complete copy
        :param kwargs:
        :return:
        """
        return self.serialize(**kwargs)

    def write_to_file(self, filename, gzip=False, complete=False, **kwargs):
        """

        :param filename:
        :param gzip:
        :param complete:
        :param kwargs: whatever is required by the subclass's serialize method
        :return:
        """
        if self._source is None:
            self._set_source(self.ref, filename)  # unless there was no source to begin with
        elif filename not in self.names:
            self._add_name(self.ref, filename)
        if complete:
            s = self._serialize_all(**kwargs)
            if self._loaded:
                s['loaded'] = True
        else:
            s = self.serialize(**kwargs)
        to_json(s, filename, gzip=gzip)



import unittest
import os

from ..lc_resource import LcResource
from ..catalog_query import READONLY_INTERFACE_TYPES
from ..archives.tests import basic_archive_src


serialized_resource = {
    'dataSource': basic_archive_src,
    'dataSourceType': 'json',
    'interfaces': list(READONLY_INTERFACE_TYPES),
    'priority': 50,
    'static': False,
    'config': {}
}


temp_file = os.path.join(os.path.dirname(__file__), 'index_file.gz')


class LcResourceTest(unittest.TestCase):
    def setUp(self):
        self.res = LcResource('test.basic', basic_archive_src, 'json',
                              interfaces=READONLY_INTERFACE_TYPES)

    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(temp_file)
        except FileNotFoundError:
            pass

    @unittest.skip('dumm')
    def test_serialize(self):
        self.assertEqual(self.res.serialize(), serialized_resource)

    def test_index_preserves_source(self):
        src = self.res.source
        self.assertIsNone(self.res.archive)
        self.res.make_index(temp_file)
        self.assertEqual(self.res.source, src)


if __name__ == '__main__':
    unittest.main()



import json
import os
from collections import defaultdict
import requests
import hashlib

from antelope import UnknownOrigin

from .archives import InterfaceError, index_archive, update_archive, create_archive

# from .foreground import LcForeground
from .catalog_query import INTERFACE_TYPES, NoCatalog, zap_inventory

# from .providers import create_archive
from . import herd_factory

class ResourceInvalid(Exception):
    """
    resource points to an invalid filename
    """
    pass


def download_file(url, local_file, md5sum=None):
    r = requests.get(url, stream=True)
    md5check = hashlib.md5()
    with open(local_file, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                md5check.update(chunk)
                # f.flush() commented by recommendation from J.F.Sebastian
    if md5sum is not None:
        assert md5check.hexdigest() == md5sum, 'MD5 checksum does not match'


class LcResource(object):
    """
    This is a record that links a semantic reference to a physical data source, and specifies the capabilities
    (and someday, access limitations) of the data source.

    The LcResource serializes to a json file with the following format:
    { ref: [ { "dataSource": source, "dataSourceType": ds_type, .... }, ... ] }
    where ref is the semantic reference.

    """
    @classmethod
    def from_archive(cls, archive, interfaces, source=None, **kwargs):
        source = source or archive.source
        ref = archive.ref
        ds_type = archive.__class__.__name__  # static flag indicates whether archive is complete
        kwargs.update(archive.init_args)
        static = kwargs.pop('static', archive.static)
        res = cls(ref, source, ds_type, interfaces=interfaces, static=static, preload_archive=archive, **kwargs)

        return res

    @classmethod
    def from_dict(cls, ref, d):
        """
        Returns a single LcResource loaded from a dict.  only required field is 'dataSourceType'.
        other fields are passed to the constructor and either interpreted directly or added as supplemental args

        If 'dataSource' is not present
        :param ref:
        :param d:
        :return:
        """
        source = d.pop('dataSource', None)
        ds_type = d.pop('dataSourceType')

        # patch to deal with changing Background extension handling
        filetype = d.pop('filetype', None)
        if filetype is not None:
            if not source.endswith(filetype):
                source += filetype

        return cls(ref, source, ds_type, **d)

    @classmethod
    def from_json(cls, file):
        """
        generates LcResources contained in the named file, sorted by increasing priority.  The filename and
        the reference must be the same.
        :param file:
        :return: an ordered list of resources
        """
        ref = os.path.basename(file)
        with open(file, 'r') as fp:
            j = json.load(fp)

        return sorted([cls.from_dict(ref, d) for d in j[ref]], key=lambda x: x.priority)

    def _instantiate(self, catalog=None):
        """
        Instantiate the archive described by the current resource.  Several steps:
         - download file if applicable
         - rename relative path to absolute path
         - instantiate the archive
         - load any cached data into the archive
         - override static spec to be consistent with archive's own spec
         - load_all() if static
         - apply_config()

        :param catalog: required for: file downloads, foreground interfaces, local paths + caches
        :return:
        """
        if self.source is None:
            # download
            if catalog is None:
                raise NoCatalog('Remote resource encountered')
            if 'download' in self._args:
                print('Downloading from %s' % self._args['download']['url'])
                self._source = catalog.download_file(localize=True, **self._args['download'])
                self.write_to_file(catalog.resource_dir)  # update resource file
            else:
                raise AttributeError('Resource has no source specified and no download information')

        if self.source.startswith('$CAT_ROOT'):
            try:
                src = catalog.abs_path(self.source)
            except AttributeError:
                raise NoCatalog('Relative path encountered but no catalog supplied')
        else:
            src = self.source

        kwargs = {**self.init_args}

        if 'foreground' in self.interfaces:
            kwargs['catalog'] = catalog

        if 0: # self.ds_type.lower() in ('foreground', 'lcforeground'):
            pass
            # self._archive = LcForeground(src, catalog=catalog, ref=self.reference, **kwargs)
        else:
            if self.ds_type.lower() == 'ecoinventlcia':
                # this is a GIANT HACK
                # we need to bring along a local ecoinvent archive to lookup flow reference qtys
                ei_ref = '.'.join(['local', 'ecoinvent', kwargs['version']])
                try:
                    res = catalog.get_resource(ei_ref, iface='exchange', strict=False)
                    res.check(catalog)
                    if hasattr(res.archive, 'load_flows'):
                        res.archive.load_flows()
                    kwargs['ei_archive'] = res.archive
                except UnknownOrigin:
                    pass

            try:
                self._archive = create_archive(src, self.ds_type, ref=self.reference,
                                               factory=herd_factory, **kwargs)
            except FileNotFoundError as e:
                raise ResourceInvalid('%s: %s' % (self.reference, e.filename))
        if catalog is not None and os.path.exists(catalog.cache_file(self.source)):
            update_archive(self._archive, catalog.cache_file(self.source))
        self._static |= self._archive.static
        if self.static and self.ds_type.lower() != 'json':
            self._archive.load_all()  # static json archives are by convention saved in complete form


    @property
    def is_loaded(self):
        return self._archive is not None

    def remove_archive(self):
        self._archive = None

    def check(self, catalog):
        if self._archive is None:
            # TODO: try/catch exceptions or return false
            # print('QQQQQQQQQQQQQQQQQQ %s QQQQQQQQQQQQQQQQQQ' % self.reference)
            self._instantiate(catalog)
            self.apply_config(catalog)  # can't remember why I set this to happen recurrently- but it's no good
        return True

    def save(self, catalog):
        self.write_to_file(catalog.resource_dir)

    def make_index(self, index_file, force=True):
        self.check(None)
        self._archive.load_all()

        the_index = index_archive(self._archive, index_file, force=force)

        return the_index

    def make_cache(self, cache_file):
        # note: do not make descendant
        self._archive.write_to_file(cache_file, complete=True, gzip=True)
        print('Created archive of %s containing:' % self._archive)
        self._archive.check_counter()

    def make_interface(self, iface):
        return self._archive.make_interface(iface)

    def apply_config(self, catalog=None):
        if len(self._config) == 0:
            return
        print('Applying stored configuration')
        if catalog is not None:
            if 'hints' in self._config:
                catalog.lcia_engine.apply_hints(self._archive.catalog_names, self._config['hints'])
        try:
            self._archive.make_interface('configure').apply_config(self._config)
        except InterfaceError:
            pass

    def add_interface(self, iface):
        iface = zap_inventory(iface)  # don't warn when interpreting resource specifications
        if iface in INTERFACE_TYPES:
            self._interfaces.add(iface)

    def _normalize_interfaces(self, interfaces):
        """
        Ensures that:
         - interfaces spec can be string or list
         - 'basic' appears
        :param interfaces:
        :return:
        """
        self.add_interface('basic')
        if interfaces is None:
            return
        if isinstance(interfaces, str):
            self.add_interface(interfaces)
        else:
            for k in interfaces:
                self.add_interface(k)

    def __init__(self, reference, source, ds_type, interfaces=None, privacy=0, priority=50, static=False,
                 preload_archive=None, **kwargs):
        """

        :param reference: semantic reference
        :param source: physical data source; 'None' allowed if 'downloadLink' argument provided
        :param ds_type: data source type
        :param interfaces: list which can include 'entity', 'foreground', or 'background'. Default 'foreground'
        :param privacy: Ignored / No longer used.
        :param priority: [50] priority level.. numeric (nominally 0-100), lowest priority resource is loaded first
        :param static: [False] if True, load_all() after initializing
        :param preload_archive: [None] use to assign an existing archive
        :param kwargs: additional keyword arguments to constructor. Some interesting ones:
          download: a dict containing 'url' and optional 'md5sum' fields
          prefix: often used when accessing zipped archives


        """
        '''
        if not os.path.exists(source):
            raise EnvironmentError('%s not found' % source)
        '''

        self._archive = preload_archive

        self._ref = reference
        if source is None:
            if 'download' not in kwargs:
                raise KeyError('Resource must be initialized with either source or download')
        self._source = source
        self._type = ds_type
        self._static = static

        self._issaved = False

        self._interfaces = set()
        self._normalize_interfaces(interfaces)

        self._priority = int(priority)

        self._internal = kwargs.pop('_internal', False)

        self._config = defaultdict(set)

        config = kwargs.pop('config', None)
        if config:
            for k, v in config.items():
                for q in v:
                    self._add_config(k, *q)

        self._args = kwargs

    def __repr__(self):
        flags = ['']
        if self.internal:
            flags.append('_int')
        if self.static:
            flags.append('static')
        if self._archive is not None:
            flags.append('loaded ')
        if len(self._config) > 0:
            flags.append('%d cfg' % len(self._config))
        fgs = ' '.join(flags)

        return 'LcResource(%s, dataSource=%s:%s, %s [%d]%s)' % (self.reference, self.source, self.ds_type,
                                                                [k for k in self.interfaces], self.priority, fgs)

    def exists(self, path):
        filename = os.path.join(path, self.reference)
        if os.path.exists(filename):
            try:
                with open(filename, 'r') as fp:
                    j = json.load(fp)
            except json.JSONDecodeError:
                return False

            if any([self.matches(k) for k in j[self.reference]]):
                return True
        return False

    @property
    def archive(self):
        return self._archive

    @property
    def is_saved(self):
        return self._issaved

    @property
    def reference(self):
        return self._ref

    @property
    def source(self):
        return self._source

    @property
    def ds_type(self):
        return self._type

    @property
    def interfaces(self):
        for k in self._interfaces:
            yield k

    @property
    def internal(self):
        return self._internal

    @property
    def priority(self):
        return self._priority

    @property
    def static(self):
        return self._static or self.ds_type.lower() == 'json'

    @property
    def init_args(self):
        return self._args

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config_dict):
        if self.archive is not None:
            try:
                cf = self.archive.make_interface('configure')
                for cfg, cfgs in config_dict.items():
                    for args in cfgs:
                        cf.check_config(cfg, args)
            except InterfaceError:
                pass
        self._config = config_dict

    def satisfies(self, ifaces):
        if ifaces is None:
            return True
        if isinstance(ifaces, str):
            ifaces = [ifaces]
        for i in ifaces:
            if i == 'basic':
                return True
            if i in self._interfaces:
                return True
        return False

    def _add_config(self, config, *args):
        """
        does no validation
        :param config:
        :param args:
        :return:
        """
        self._config[config].add(args)

    def _serialize_config(self):
        j = dict()
        for k, v in self._config.items():
            j[k] = sorted([list(g) for g in v], key=lambda x: x[0])
        return j

    def export_config(self):
        return self._serialize_config()

    def serialize(self, stripped=False):
        if stripped:
            j = {"static": self.static,
                 "priority": self.priority}
        else:
            j = {
                "dataSource": self.source,
                "dataSourceType": self.ds_type,
                "interfaces": sorted([k for k in self.interfaces]),
                "priority": self.priority,
                "static": self.static
            }
        j.update(self._args)
        if self.internal:
            j['_internal'] = True
        j['config'] = self._serialize_config()
        return j

    def matches(self, k):
        """
        Pretty cheesy.  When we serialize a set of resources, we need to make sure not to include self twice.  To
        make the comparison concrete, use a serialized resource as input.

        We were using dataSource as a unique identifier for resource entries; but the introduction of download links
         breaks that because a downloadable resource has no source until it's been downloaded.
         The solution is to fallback to download.url ONLY IF the resource has no source specified.
        :param k: a serialized LcResource
        :return:
        """
        if k['dataSource'] is not None and self.source is not None:
            return k['dataSource'] == self.source
        return k['download']['url'] == self._args['download']['url']

    def write_to_file(self, path, assign_ref=None, apply_config=None):
        """
        Adds the resource to a file whose name is the resource's semantic reference. If the same datasource is
        already present in the file, replace it with the current resource.  otherwise append.
        :param path: directory to store the resource file.
        :param assign_ref: assign this ref instead of the resource's current ref
        :param apply_config: overwrites configuration with supplied dict
        :return:
        """
        if assign_ref is None:
            assign_ref = self.reference
        if apply_config is not None:
            self.config = apply_config  # tests configuration before storing it
        if not os.path.isdir(path):
            if os.path.exists(path):
                raise ValueError('Please provide a directory path')
            os.makedirs(path)


        filename = os.path.join(path, assign_ref)
        if os.path.exists(filename):
            try:
                with open(filename, 'r') as fp:
                    j = json.load(fp)
            except json.JSONDecodeError:
                j = {assign_ref: []}

            resources = [k for k in j[assign_ref] if not self.matches(k)]
            resources.append(self.serialize())
        else:
            resources = [self.serialize()]
        with open(os.path.join(path, assign_ref), 'w') as fp:
            json.dump({assign_ref: resources}, fp, indent=2)
        self._issaved = True



from __future__ import print_function, unicode_literals
from numbers import Number
from math import isclose

import uuid

from collections import defaultdict

from antelope import check_direction

from .entities import LcEntity
from ..exchanges import ExchangeValue, DuplicateExchangeError, AmbiguousReferenceError


class ZeroAllocation(Exception):
    pass


class MissingAllocation(Exception):
    pass


class NoExchangeFound(Exception):
    pass


class MultipleReferencesFound(Exception):
    """
    Whereas AmbiguousReferenceError indicates that further filtering is possible; MultipleReferencesFound indicates
    that there is no way to provide additional information.
    """
    pass


class NotAReference(Exception):
    pass


class AlreadyAReference(Exception):
    pass


class ReferenceSettingFailed(Exception):
    pass


class LcProcess(LcEntity):

    _ref_field = 'referenceExchange'
    _new_fields = ['SpatialScope', 'TemporalScope', 'Classifications']

    @classmethod
    def new(cls, name, **kwargs):
        """
        :param name: the name of the process
        :return:
        """
        u = uuid.uuid4()
        return cls(str(u), entity_uuid=u, Name=name, **kwargs)

    def __init__(self, external_ref, **kwargs):
        """
        THe process's data is a set of exchanges.

        A process's reference entity is a subset of these.  It is an error for these exchanges to have terminations
        (if they're terminated, they're not reference flows- they're dependencies). These references can be used
        as allocation keys for the exchanges.

        The entities in reference_entity and _exchanges are not necessarily the same, although they should hash the
        same.  Not sure whether this is a design flaw or not- but the important thing is that reference entities do
        not need to have exchange values associated with them (although they could).

        process.find_reference(key), references() [generator], and reference(flow) all return entries from _exchanges,
        not entries from reference_entity.  The only public interface to the objects in reference_entity is
        reference_entity itself.
        :param entity_uuid:
        :param kwargs:
        """
        self._exchanges = dict()  # maps exchange key to exchange
        self._exch_map = defaultdict(set)  # maps flow external_ref to exchanges having that flow

        super(LcProcess, self).__init__('process', external_ref, **kwargs)
        if self._reference_entity is not None:
            raise AttributeError('How could the reference entity not be None?')
        self._reference_entity = dict()  # it is not possible to specify a valid reference_entity on init
        self._alloc_by_quantity = None
        self._alloc_sum = 0.0

        if 'SpatialScope' not in self._d:
            self._d['SpatialScope'] = 'GLO'
        if 'TemporalScope' not in self._d:
            self._d['TemporalScope'] = '0'
        if 'Classifications' not in self._d:
            self._d['Classifications'] = []

    @property
    def name(self):
        return self['Name']

    def __str__(self):
        return '%s [%s]' % (self._d['Name'], self._d['SpatialScope'])

    def __len__(self):
        return len(self._exchanges) + len(self._reference_entity)

    @property
    def reference_entity(self):
        return set(self._reference_entity.values())

    def _validate_reference(self, ref_set):
        """
        An exchange may be a reference only if it is null-terminated or terminated to a non-elementary context.
        :param ref_set:
        :return:
        """
        for x in ref_set:
            if x.termination is not None and x.type not in ('context', 'reference'):  # excludes 'node', 'self'
                return False
            if not super(LcProcess, self)._validate_reference(x):
                return False
        return True

    def _print_ref_field(self):
        return 'see exchanges'

    def _set_reference(self, exch_to_be_ref):
        """
        :param exch_to_be_ref: a dependent exchange to be made a reference exchange. This will fail if the exchange
        is already a reference.
        :return:
        """
        if not self._validate_reference({exch_to_be_ref}):
            raise ReferenceSettingFailed(exch_to_be_ref)

        if exch_to_be_ref.key in self._exchanges:
            assert exch_to_be_ref is self._exchanges[exch_to_be_ref.key]
            self._exchanges.pop(exch_to_be_ref.key)
            if exch_to_be_ref.set_ref(self):
                self._reference_entity[exch_to_be_ref.flow.external_ref, exch_to_be_ref.direction] = exch_to_be_ref
            else:
                raise ReferenceSettingFailed('%s\n%s' % (self, exch_to_be_ref))
        else:
            raise NoExchangeFound
        if self.alloc_qty is not None:
            self.allocate_by_quantity(self._alloc_by_quantity)

    @property
    def alloc_qty(self):
        if self._alloc_sum != 0:
            return self._alloc_by_quantity
        return None

    @property
    def alloc_total(self):
        return self._alloc_sum

    def show_inventory(self, reference=None):
        """
        Convenience wrapper around self.inventory() which:
         * sorts the exchanges by reference, then by direction
         * prints the exchanges to output
         * provides an enumeration of exchanges for interactive access
         = returns the exchanges as a sorted list.
        :param reference:
        :return:
        """
        num = 0
        it = sorted(self.inventory(reference), key=lambda x: (not x.is_reference, x.direction))
        if reference is None:
            print('%s' % self)
        else:
            print('Reference: %s' % reference)
        for i in it:
            print('%2d %s' % (num, i))
            num += 1
        return it

    def _gen_exchanges(self, flow=None, direction=None, reference=None):
        """
        Generate a list of exchanges matching the supplied flow and direction.

        If this is too slow, we could hash the process's exchanges by flow
        :param flow: either a flow entity, flow ref, or external_ref
        :param direction: [None] or filter by direction
        :param reference: [None] find any exchange; True: only reference exchanges; False: only non-reference exchanges
        :return:
        """
        if flow is None:
            if reference is True:
                _x_gen = self.references()
            elif reference is False:
                _x_gen = (x for x in self._exchanges.values())
            else:
                _x_gen = (x for s in self._exch_map.values() for x in s)
        else:
            if hasattr(flow, 'entity_type'):
                if flow.entity_type != 'flow':
                    raise TypeError('Flow argument must be a flow')
                flow = flow.external_ref
            if reference is True:
                _x_gen = (x for x in self._exch_map[flow] if x.is_reference)
            elif reference is False:
                _x_gen = (x for x in self._exch_map[flow] if not x.is_reference)
            else:
                _x_gen = (x for x in self._exch_map[flow])
        for x in _x_gen:
            if direction is not None:
                if x.direction != direction:
                    continue
            yield x

    def get_exchange(self, key):
        try:
            return self._exchanges[key]
        except KeyError:
            try:
                return next(x for x in self.reference_entity if x.key == key)
            except StopIteration:
                raise KeyError

    def exchanges(self, flow=None, direction=None):
        for x in self._gen_exchanges(flow=flow, direction=direction):
            yield x.trim()

    def exchange_values(self, flow, direction=None):
        """
        Yield full exchanges matching flow specification.  Flow specification required.
        Will only yield multiple results if there are multiple terminations for the same flow.
        :param flow:
        :param direction:
        :return:
        """
        for x in self._gen_exchanges(flow=flow, direction=direction):
            yield x

    def has_exchange(self, flow, direction=None):
        try:
            next(self.exchange_values(flow, direction=direction))
        except StopIteration:
            return False
        return True

    def inventory(self, ref_flow=None):
        """
        generate a process's exchanges.  If no reference is supplied, generate unallocated exchanges, including all
        reference exchanges.  If a reference is supplied AND the process is allocated with respect to that reference,
        generate ExchangeValues as allocated to that reference flow, and exclude reference exchanges.  If a reference
        is supplied but the process is NOT allocated to that reference, generate unallocated ExchangeValues (excluding
        the reference itself).  Reference must be a flow or exchange found in the process's reference entity.

        :param ref_flow:
        :return:
        """
        if ref_flow is None:
            ref_exch = None
            for i in self.reference_entity:
                yield i
        else:
            ref_exch = self.reference(ref_flow)
        for i in self._exchanges.values():
            if ref_exch is None:
                # generate unallocated exchanges
                yield i
            elif ref_exch.is_reference:
                # generate allocated, normalized, non-reference exchanges
                if i in self.reference_entity:
                    continue
                else:
                    yield ExchangeValue.from_allocated(i, ref_exch)
            else:
                # generate un-allocated, normalized, non-query exchanges
                if i is ref_exch:
                    continue
                else:
                    yield ExchangeValue.from_allocated(i, ref_exch)

    def exchange_relation(self, ref_flow, exch_flow, direction, termination=None):
        rx = self.reference(ref_flow)
        if termination is None:
            xs = [x for x in self.exchange_values(flow=exch_flow, direction=direction)]
            if len(xs) == 1:
                return xs[0].value / rx.value
            elif len(xs) == 0:
                return 0.0
            else:
                return sum([x.value for x in xs]) / rx.value
        else:
            x = self.get_exchange(hash((self.external_ref, exch_flow, direction, termination)))
            return x[rx]

    def find_reference(self, spec=None, direction=None):
        """
        returns a reference exchange matching the specification.

        If multiple results are found, filters out terminated exchanges

        :param spec: could be None, external_ref, flow, flow ref, or exchange
        :param direction: could be helpful if the object is a non-reference exchange
        :return:
        """
        if hasattr(spec, 'entity_type'):
            if spec.entity_type == 'exchange':
                if direction is None:
                    direction = spec.direction
                _x_gen = self._gen_exchanges(flow=spec.flow.external_ref, direction=direction, reference=True)
            elif spec.entity_type == 'flow':
                _x_gen = self._gen_exchanges(flow=spec.external_ref, direction=direction, reference=True)
            else:
                raise TypeError('Cannot interpret specification %s (%s)' % (spec, type(spec)))
        else:  # works for spec=external_ref or spec is None
            _x_gen = self._gen_exchanges(spec, direction=direction, reference=True)

        candidates = list(_x_gen)

        if len(candidates) == 0:
            raise NoExchangeFound('%s: %s' % (self.external_ref, spec))
        elif len(candidates) > 1:
            raise MultipleReferencesFound('%d exchanges found; try specifying direction' % len(candidates))
        else:
            return candidates[0]

    '''
    def _strip_term(self, flow, dirn):
        """
        Removes an existing terminated exchange and replaces it with an unterminated one
        """
        exs = [k for k in self._gen_exchanges(flow, dirn) if k.termination is not None]
        if len(exs) > 1:
            raise DuplicateExchangeError('%d terminated exchanges found for %s: %s' % (len(exs), dirn, flow))
        elif len(exs) == 0:
            return  # NoExchangeFound in _set_reference instead of here
        ex = exs[0]

        old = self._exchanges.pop(ex.key)
        new = old.reterminate(None)
        if new.key in self._exchanges:
            raise KeyError('Unterminated exchange already exists!')
        self._exchanges[new.key] = new
        self._exch_map[new.flow.external_ref].remove(old)
        self._exch_map[new.flow.external_ref].add(new)
    '''

    def set_reference(self, flow, dirn):
        """
        Exchange must already exist.
        fmr: If the exchange is currently terminated, the termination is removed.
        now: exchanges terminated to non-elementary context are now allowed
        :param flow:
        :param dirn:
        :return:
        """
        # self._strip_term(flow, dirn)
        dirn = check_direction(dirn)
        if (flow.external_ref, dirn) in self._reference_entity:
            # already a reference
            return self._reference_entity[flow.external_ref, dirn]
        else:
            rx = list(self._gen_exchanges(flow, dirn, reference=False))
            if len(rx) == 0:
                raise NoExchangeFound(flow, dirn)
            elif len(rx) > 1:
                raise AmbiguousReferenceError(rx)
            else:
                self._set_reference(rx[0])
                return rx[0]

    def remove_reference(self, flow, dirn):
        dirn = check_direction(dirn)
        k = (flow.external_ref, dirn)
        rx = self._reference_entity[k]
        if rx.key in self._exchanges:
            raise DuplicateExchangeError(rx)
        self._reference_entity.pop(k)
        rx.unset_ref(self)
        self.remove_allocation(rx)
        self._exchanges[rx.key] = rx
        if self._alloc_by_quantity is not None:
            self.allocate_by_quantity(self._alloc_by_quantity)

    def references(self):
        for rf in self.reference_entity:
            yield rf

    def reference(self, flow_ref=None):
        if flow_ref in self.reference_entity:
            return flow_ref
        return self.find_reference(flow_ref)

    ''' # don't think I want this
    def reference_value(self, flow=None):
        return self.get_exchange(self.find_reference(flow).key).value
    '''

    def has_reference(self, flow=None):
        try:
            self.find_reference(flow)
            return True
        except NoExchangeFound:
            return False

    def _alloc_dict(self, quantity=None):
        """
        Returns a dict mapping reference key to NON-NORMALIZED allocation factor.  This factor reports the amount of
        the reference flow, in dimensions of the specified quantity, that is produced by a UNIT activity of the process.
        Normalized allocation factors are obtained by dividing by the sum of these amounts.
        :param quantity:
        :return:
        """
        if quantity is None:
            if self._alloc_by_quantity is None:
                raise ValueError('An allocation quantity is required to compute normalized allocation factors')
            quantity = self._alloc_by_quantity

        return {rf: rf.value * quantity.cf(rf.flow)
                for rf in self.reference_entity
                if rf.value is not None}

    def allocate_by_quantity(self, quantity):
        """
        Store a quantity for partitioning allocation.  All non-reference exchanges will have their exchange values
        computed based on the total, determined by the quantity specified.  For each
        reference exchange, computes the magnitude of the quantity output from the unallocated process. Reference flows
        lacking characterization in that quantity will receive zero allocation.

        Each magnitude is the allocation numerator for that reference, and the sum of the magnitudes is the allocation
        denominator.
        :param quantity: an LcQuantity (or None to remove quantity allocation)
        :return:
        """
        if quantity is None:
            self._alloc_by_quantity = None
            self._alloc_sum = 0.0
            self._d.pop('AllocatedByQuantity', None)
            return

        mags = self._alloc_dict(quantity)

        total = sum([v for v in mags.values()])
        if total == 0:
            raise ZeroAllocation('%s: zero total found; not setting allocation by qty %s' % (self.external_ref, quantity))

        self._alloc_by_quantity = quantity
        self._alloc_sum = total
        self['AllocatedByQuantity'] = quantity

    def allocation_factors(self, quantity=None):
        """
        Returns a dict mapping reference exchange to that reference's allocation factor according to the specified
        allocation quantity.
        If no quantity is specified, the current allocation quantity is used.  DOES NOT AFFECT CURRENT ALLOCATION.
        :param quantity: allocation quantity
        :return:
        """
        d = self._alloc_dict(quantity)
        s = sum(d.values())
        return {k: v / s for k, v in d.items()}

    def is_allocated(self, reference, strict=False):
        """
        Tests whether a process's exchanges contain allocation factors for a given reference.
        :param reference:
        :param strict: [False] if True, raise an exception if some (but not all) exchanges are missing allocations.
        :return: True - allocations exist; False - no allocations exist; raise MissingFactor - some allocations exist
        """
        try:
            reference = self.find_reference(reference)
        except NoExchangeFound:
            print('Not a reference exchange.')
            return False
        if self.alloc_qty is not None:
            if self.alloc_qty.cf(reference.flow) == 0:
                return False
            return True
        missing_allocations = []
        has_allocation = []
        for x in self._exchanges.values():
            if x.is_allocated(reference):
                has_allocation.append(x)
            else:
                missing_allocations.append(x)
            if not strict:
                if len(has_allocation) > 0:
                    return True  # for nonstrict, bail out as soon as any allocation is detected
        if len(has_allocation) * len(missing_allocations) == 0:
            if len(has_allocation) == 0:
                return False
            return True
        if strict:
            for x in missing_allocations:
                print('in process %s [%s]\nReference: %s' % (self['Name'], self.external_ref,
                                                             reference.flow.external_ref))
                print('%s' % x)
                raise MissingAllocation('Missing allocation factors for above exchanges')

    def test_allocation_consistency(self, flow=None, display=True):
        """
        For each non-reference item in the inventory, test that the values allocated to each reference, weighted by the
        reference values, sum up to the un-allocated value.

        :param flow: [None] single flow to test; defaults to entire inventory
        :param display: [True] whether to print output to the screen
        :return:
        """
        _d = {1: '', 0: 'X'}
        if flow is None:
            xs = list(self._exchanges.values())
        else:
            xs = list(self._gen_exchanges(flow))
        if display:
            print('%-60.60s %-10.10s %-10.10s' % ('exchange', 'total', 'unalloc'))
        test = True
        for x in xs:
            total = sum(x[rx] * rx.value for rx in self.references())
            loc = isclose(total, x.value, rel_tol=1e-8)
            test &= loc
            if display:
                print('%-60.60s %10.10s %10.10s  %s' % (x, total, x.value, _d[int(loc)]))
        if display:
            print({True: 'PASS', False: 'FAIL'}[test])
        return test

    def remove_allocation(self, reference):
        for x in self._exchanges.values():
            x.remove_allocation(reference)

    def unobserved_lci(self, *args, **kwargs):
        raise NotImplementedError

    def add_exchange(self, flow, dirn, reference=None, value=None, termination=None, add_dups=False):
        """
        This is used to create Exchanges and ExchangeValues and AllocatedExchanges.

        If the flow+dir+term is already in the exchange set:
            if no reference is specified and/or no value is specified- nothing to do
            otherwise (if reference and value are specified):
                upgrade the exchange to an allocatedExchange and add the new reference exch val
        otherwise:
            if reference is specified, create an AllocatedExchange
            otherwise create an Exchange / ExchangeValue

        :param flow:
        :param dirn:
        :param reference:
        :param value:
        :param termination: None for reference or cutoff flows; a context for elementary flows; a valid external_ref
         for terminated intermediate flows.
        :param add_dups: (False) set to true to handle "duplicate exchange" errors by cumulating their values
        :return:
        """
        dirn = check_direction(dirn)
        if (flow.external_ref, dirn) in self._reference_entity:
            raise AlreadyAReference((flow, dirn))
        _x = hash((self.external_ref, flow.external_ref, dirn, termination))
        if _x in self._exchanges:
            if value is None or value == 0:
                return None
            e = self._exchanges[_x]
            if reference is None:
                if isinstance(value, dict):
                    e.update(value)
                else:
                    try:
                        e.value = value  # this will catch already-set errors
                    except DuplicateExchangeError:
                        if add_dups:
                            e.add_to_value(value)
                        else:
                            print('Duplicate exchange in process %s:\n%s' % (self.external_ref, e))
                            raise
                return e

            else:
                try:
                    e[reference] = value  # this will catch already-set errors
                except DuplicateExchangeError:
                    if add_dups:
                        e.add_to_value(value, reference=reference)
                    else:
                        print('Duplicate exchange in process %s:\n%s' % (self.external_ref, e))
                        raise
                except ValueError:
                    print('Error adding [%s] = %10.3g for exchange\n%s\nto process\n%s' % (
                        reference.flow.external_ref, value, e, self.external_ref))
                    raise

                return e

        else:
            if isinstance(value, Number) or value is None:
                if reference is None:
                    e = ExchangeValue(self, flow, dirn, value=value, termination=termination)
                else:
                    if reference not in self.reference_entity:
                        raise KeyError('Specified reference is not registered with process: %s' % reference)
                    e = ExchangeValue(self, flow, dirn, value=None, termination=termination)
                    e[reference] = value

            elif isinstance(value, dict):
                e = ExchangeValue(self, flow, dirn, value_dict=value, termination=termination)
            else:
                raise TypeError('Unhandled value type %s' % type(value))

            # This is the only point an exchange (must be ExchangeValue) is added to the process (see also _strip_term)
            self._exchanges[e.key] = e
            self._exch_map[e.flow.external_ref].add(e)
            return e

    def merge(self, other):
        raise NotImplemented('This should be done via fragment construction + aggregation')

    def serialize(self, exchanges=False, domesticate=False, drop_fields=(), **kwargs):
        j = super(LcProcess, self).serialize(domesticate=domesticate, drop_fields=drop_fields)
        j.pop(self._ref_field)  # reference reported in exchanges
        j['exchanges'] = sorted([x.serialize(**kwargs) for x in self.reference_entity],
                                key=lambda x: (x['direction'], x['flow']))
        if exchanges:
            # if exchanges is true, report all exchanges
            j['exchanges'] += sorted([x.serialize(**kwargs) for x in self._exchanges.values()],
                                     key=lambda x: (x['direction'], x['flow']))
        else:
            j.pop('allocationFactors', None)  # added just for OpenLCA JSON-LD, but could be generalized
        return j



from synonym_dict import SynonymDict, SynonymSet


class QuantityUnitMismatch(Exception):
    pass


class QuantityAlreadySet(Exception):
    pass


class ValidationError(Exception):
    pass


class QuantitySynonyms(SynonymSet):
    """
    QuantitySynonyms are string terms that all refer to the same quantity of measure. They must all have the same
    unit, because they are used to define the unit of measure of flowables.  To repeat: quantity instances that have
    the same dimensionality but different units (e.g. kWh and MJ) are NOT SYNONYMS but distinct quantities. The
    LciaEngine should be able to handle conversions between these kinds of quantities.
    """
    @classmethod
    def new(cls, quantity):
        return cls(*quantity.quantity_terms(), quantity=quantity)

    def _validate_quantity(self, quantity):
        if not hasattr(quantity, 'entity_type'):
            raise AttributeError(quantity)
        if quantity.entity_type != 'quantity':
            raise TypeError(quantity)
        if not hasattr(quantity, 'unit'):
            raise AttributeError(quantity, 'unit')
        if quantity.unit is not None:  # allow unit=None quantities to merge with existing quantities
            if not isinstance(quantity.unit, str):
                raise AttributeError(quantity, 'unit-str')
            if self.unit is not None and quantity.unit != self.unit:
                raise QuantityUnitMismatch('incoming %s (set %s)' % (quantity.unit, self.unit))
        return True

    def __init__(self, *args, quantity=None, unit=None, **kwargs):
        self._quantity = None
        super(QuantitySynonyms, self).__init__(*args, **kwargs)
        self._unit = unit
        if quantity is not None:
            self.quantity = quantity

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, item):
        if self.quantity is None:
            if self._validate_quantity(item):
                self._quantity = item
                self._unit = item.unit
                for term in item.quantity_terms():
                    self.add_term(term)
            else:
                raise ValidationError('Quantity fails validation for uncaught reason (%s)' % item)
        else:
            raise QuantityAlreadySet

    def _save_synonyms(self, other):
        """
        adds to other's synonym set-
        :param other:
        :return:
        """
        for k in other.terms:
            self._quantity.add_synonym(k)

    def add_child(self, other, force=False):
        if not isinstance(other, QuantitySynonyms):
            raise TypeError('Child set is not a Quantity synonym set (%s)' % type(other))
        if other.unit is not None:
            if self.unit is None:
                self._unit = other.unit
            elif other.unit != self.unit:
                raise QuantityUnitMismatch('incoming %s (canonical %s)' % (other.unit, self._quantity.unit))
        if self._quantity is None:
            self.quantity = other.quantity
        elif self._quantity.is_entity:
            self._save_synonyms(other)
        super(QuantitySynonyms, self).add_child(other, force=force)

    @property
    def object(self):
        return self._quantity

    @property
    def unit(self):
        return self._unit

    def serialize(self):
        d = super(QuantitySynonyms, self).serialize()
        if self.unit is not None:
            d['unit'] = self.unit

        return d


class QuantityManager(SynonymDict):

    _entry_group = 'Quantities'
    _syn_type = QuantitySynonyms

    _ignore_case = True

    def _add_from_dict(self, j):
        name = j['name']
        syns = j.pop('synonyms', [])
        unit = j.pop('unit', None)
        self.new_entry(name, *syns, unit=unit, merge=False)

    def find_matching_quantity(self, quantity):
        if isinstance(quantity, str):
            return self[quantity]
        for term in quantity.quantity_terms():
            if term in self:
                return self[term]
        raise KeyError(quantity)

    def add_quantity(self, quantity):
        """
        Prunes terms on quantity unit mismatch
        :param quantity:
        :return:
        """
        new_q = QuantitySynonyms.new(quantity)
        try:
            self.add_or_update_entry(new_q, merge=True, create_child=True)
        except QuantityUnitMismatch:
            # want to print a warning
            # if we get QuantityUnitMismatch, then that means we didn't get a MergeError
            me = self.match_entry(*new_q.terms)
            print('!! Warning: new quantity %s [%s] has unit conflict with existing quantity %s [%s] ' % (new_q,
                                                                                                          new_q.unit,
                                                                                                          me, me.unit))
            pe = self.add_or_update_entry(new_q, merge=False, prune=True)
            print('   Added pruned quantity %s [%s]' % (pe.name, pe.unit))

    def add_synonym(self, term, syn):
        super(QuantityManager, self).add_synonym(term, syn)
        ent = self._d[syn]
        # canonical quantity is always a ref
        for child in ent.children:
            if child.quantity is not None and child.quantity.is_entity:
                child.quantity.add_synonym(str(syn).strip())

    def __getitem__(self, item):
        if hasattr(item, 'link'):
            item = item.link
        return super(QuantityManager, self).__getitem__(item)



"""
Class for interchanging entities with an excel spreadsheet.

There are two basic parts to this operation:
 * we want to be able to write a subset of entities in an archive to an excel worksheet whose name is the same as the
   entity_type. We do not want to mix different entity types in the same worksheet. We also only want to include
   metadata-- quantitative data (exchanges and flow characterizations) are not meant to be manipulated in this way.
   The export to Excel should include the required columns to the leftmost, followed by nonrequired properties.  For a
   given entity/property pair, if the entity lacks that property the entry should be blank.

 * We want to be able to read in metadata from a spreadsheet as well- creating new entities as needed, and modifying
   existing entities nondestructively according to a policy with two options:
    - 'defer': if a property exists for a given entity, let its current value stand. If a property does not exist and
      the incoming specification is non-null, assign it..
    - 'overwrite': Assign all incoming non-null specifications

For the first one, the only capability of the archive that is needed is (arguably) external to the feature-- namely,
the ability to generate the entities. Therefore I think it should be a utility function and not an object at all.

For the second one here, this seems like more a capability of an archive rather than a separate class. I think this
should be a context mgr.  We (for now) want to be strict about the sheet naming- let's take that flexibility away
from the user.

The required fields should be: external_ref, *signature_fields.  uuid is supported but not required
"""
import os
import re
import xlrd
from synonym_dict.lower_dict import LowerDict

from antelope import EntityNotFound
from ..implementations.quantity import convert
from ..characterizations import DuplicateCharacterizationError


class XlsxEntityTypeWriter(object):
    _attribs = ('external_ref', 'uuid')

    def _next_name(self):
        name = self._etype
        k = 0
        while name in self._wb.sheetnames:
            name = '%s%d' % (self._etype, k)
            k += 1
        print('Writing entities to sheet %s' % name)
        return name

    def _add_column(self, col_name):
        if col_name not in self._rc:
            i = len(self._columns)
            self._columns.append(col_name)
            self._rc[col_name] = i
            self._sheet.write(0, i, col_name)
        return self._rc[col_name]

    def __init__(self, xls_workbook, etype):
        self._wb = xls_workbook
        self._etype = etype
        self._columns = []
        self._rc = LowerDict()
        self._count = 0
        self._sheet = self._wb.add_worksheet(self._next_name())
        self._sig = None
        self._ref_col = None

    def _grab_signature_fields(self, entity):
        if self._sig:
            return
        for k in self._attribs:
            self._add_column(k)
        for k in entity.signature_fields():
            i = self._add_column(k)
            # tag the reference column
            if k == entity.reference_field:
                self._ref_col = i
        self._sig = tuple(self._columns)

    def _write(self, row, column, value):
        """
        For other entities
        For non-writeable types, deliver their __repr__, preceded by an asterisk
        """
        if hasattr(value, 'entity_type'):
            value = '!%s' % value.external_ref
        elif isinstance(value, str) and (value.startswith('*') or value.startswith('!')):
            value = '*%s' % repr(value)
        try:
            self._sheet.write(row, column, value)
        except TypeError:
            self._sheet.write(row, column, '*%s' % repr(value))

    def add_entity(self, entity):
        if entity.entity_type != self._etype:
            print('Skipping entity %s of type %s' % (entity.external_ref, entity.entity_type))
            return False
        if self._sig is None:
            self._grab_signature_fields(entity)

        self._count += 1

        # first - signature fields
        for i, col in enumerate(self._sig):
            if col in self._attribs:
                val = getattr(entity, col)
            elif i == self._ref_col and self._etype == 'process':
                val = {repr(k) for k in entity.reference_entity}
            else:
                val = entity[col]
            self._write(self._count, i, val)

        # next - other properties
        p = set(entity.properties()) - set(self._sig)
        for k in p:
            i = self._add_column(k)
            self._write(self._count, i, entity[k])
        return True


def write_to_excel(filename, entity_iter, overwrite=False):
    if not bool(re.search('xlsx?$', filename, flags=re.I)):
        filename += '.xlsx'
    if os.path.exists(filename):
        if not overwrite:
            raise FileExistsError('Use overwrite=True')

    from xlsxwriter import Workbook

    with Workbook(filename) as w:
        count = 0
        d = dict()
        for ent in entity_iter:
            etype = ent.entity_type
            if etype not in d:
                d[etype] = XlsxEntityTypeWriter(w, etype)
            sh = d[etype]
            if sh.add_entity(ent):
                count += 1
        print('Writing %d entities to %s' % (count, filename))
        # w.close()  # happens automatically with context exit


def _check_merge(merge):
    if merge in ('defer', 'overwrite'):
        return merge
    raise ValueError('Invalid merge strategy: %s' % merge)


class XlsxUpdater(object):
    """
    This class uses the contents of a properly formatted XLS file (or XLS-file-like object) to create or update
    entities in an archive.

    Uses sheet names that match the archive's "_entity_types" property, so 'flow' and 'quantity' nominally.  'flows' 
    and 'quantities' also work.

    Each sheet must be in strict tabular format, with the first row being headers and each subsequent row being one
    entity.

    The only required column is 'external_ref', though 'name' and 'reference' are recommended for new entities.  NOTE:
    for the time being, quantities require a 'referenceUnit' column and flows require a 'referenceQuantity' column,
    though the hope is to eliminate that requirement in the future. For quantities, the 'referenceUnit' column should
    be a unit string; for flows the 'referenceQuantity' column should be a known quantity signifier (external_ref,
    link, or canonical name recognized by the term manager)

    Optional columns include 'uuid' and 'origin'

    All other columns are assigned as properties.

    FLOW PROPERTIES
    If there is a sheet called 'flowproperties', then it is interpreted as a list of flow characterizations.  This
    sheet can have the following columns:
     - 'flow'
     - 'ref_quantity' - (optional) if present, used only as a check against the flow's native reference_entity
     - 'ref_unit' - (optional) a convertible unit in ref_quantity
     - 'quantity' - the quantity being characterized
     - 'unit' - (optional) a convertible unit in quantity
     - 'value' - the characterization factor
     - 'context' - (optional) context for the characterization

    CELL CONTENTS
    Cells are read as strings, except for the following:
     - if a cell's first character is '!', the subsequent characters are interpreted as an entity reference
     - if a cell's first character is '*', the subsequent characters are EVALUATED, so obviously this is terribly
       insecure, but it was intended to allow people to store lists, sets, and dicts.

    MERGE STRATEGY
    The updater has two merge strategies:
      * "defer": (default) If an entity exists and has a property already defined, defer to the existing property
      * "overwrite": Any non-null incoming property is assigned to the entity, overwriting any existing value.
    """
    def __init__(self, xlrd_like, merge='defer', quiet=True):
        """
        Updates entity meta-information from a spreadsheet created from above. external_ref is used as a key;
        if an entity is not found, a new one is created.  'uuid' and reference fields are only used for creation
        of new entities.
        Note: uses the default sheet names of 'flow' and 'quantity'

        :param xlrd_like: XLRD workbook-like object OR filename of XLS file
        :param merge: ['defer'] - do not overwrite existing properties
                      ['overwrite'] - do overwrite existing properties
        :return:
        """
        self._quiet = quiet
        self._merge = _check_merge(merge)
        if isinstance(xlrd_like, str) and os.path.exists(xlrd_like):
            self._xl = xlrd.open_workbook(xlrd_like)
        else:
            self._xl = xlrd_like

    @property
    def ar(self):
        return NotImplemented

    @property
    def origin(self):
        return NotImplemented

    @property
    def qi(self):
        return NotImplemented

    def _new_entity(self, etype, rowdata):
        raise NotImplementedError

    def apply(self):
        raise NotImplementedError

    def get_flow(self, flow):
        return NotImplemented

    def get_context(self, cx):
        return NotImplemented

    def _grab_value(self, cell):
        value = cell.value
        if isinstance(value, str):
            if value.startswith('*'):
                value = eval(value[1:])
            elif value.startswith('!'):
                value = self.ar[value[1:]]
            elif value == '':
                value = None
        return value

    def _process_flow_properties(self):
        fp, headers = self._sheet_accessor('flowproperties')
        if fp is None:
            return

        for row in range(1, fp.nrows):
            rowdata = {headers[i]: self._grab_value(k) for i, k in enumerate(fp.row(row)[:len(headers)])}
            try:
                flow = self.get_flow(rowdata['flow'])
            except EntityNotFound:
                self._print('Skipping unknown flow %s' % rowdata['flow'])
                continue
            rq_spec = rowdata.pop('ref_quantity', None)
            if rq_spec is not None:
                try:
                    rq = self.qi.get_canonical(rq_spec)
                except EntityNotFound:
                    print('%s Skipping record with invalid ref quantity %s' % (rowdata['flow'], rq_spec))
                    continue
                if rq != flow.reference_entity:
                    print('%s ref quantity (%s) does not agree with spec %s: skipping cf' % (rowdata['flow'],
                                                                                             flow.reference_entity,
                                                                                             rq))
                    continue
            else:
                rq = flow.reference_entity

            try:
                qq = self.qi.get_canonical(rowdata['quantity'])
            except EntityNotFound:
                print('Skipping unrecognized canonical quantity %s' % rowdata['quantity'])
                continue

            value = rowdata.pop('value', None)

            cx = rowdata.pop('context', None)
            if cx is not None:
                cx = self.get_context(cx)
                # raise NotImplementedError('TODO contexts!')

            loc = rowdata.pop('location', None) or rowdata.pop('locale', None) or 'GLO'

            if value is None:
                continue
            try:
                value = float(value)
            except (TypeError, ValueError):
                print('Skipping non-numeric value entry %s' % value)
                continue

            refunit = rowdata.pop('ref_unit', None)
            if refunit is not None:
                try:
                    value *= convert(rq, to=refunit)
                except KeyError as e:
                    print(e.args)
                    continue

            unit = rowdata.pop('unit', None)
            if unit is not None:
                try:
                    value *= convert(qq, from_unit=unit)
                except KeyError as e:
                    print(e.args)
                    continue

            if self._merge == 'overwrite':
                flow.characterize(qq, value=value, context=cx, overwrite=True, origin=self.origin, location=loc)
                self._print('Characterizing %s: %g %s / %s' % (flow, value, qq.unit, rq.unit))
            else:
                try:
                    flow.characterize(qq, value=value, context=cx)
                    self._print('Characterizing %s: %g %s / %s' % (flow, value, qq.unit, rq.unit))
                except DuplicateCharacterizationError:
                    self._print('Deferring to existing CF')
                    continue

    _vn = {'flow': 'flows',
           'quantity': 'quantities'}

    def _sheet_accessor(self, sheetname):
        """
        Returns a 2-tuple of sheet, headers
        'sheet' must be an "xlrd.Sheet-like" object that implements the following very simple API:
         sheet.nrows - returns number of rows
         sheet.row(n) - returns the contents of the nth row in a list of "cell-like" objects
        'cell-like' object has 2 properties:
         'ctype' - integer 0=empty, 1=text, 2=number, 3=date, 4=bool, 5=error, 6=debug
         'value' - contents of the cell
        'headers' is the contents of the 0th row in a list
        :param sheetname:
        :return: sheet, headers
        """
        if sheetname in self._xl.sheet_names():
            sh = self._xl.sheet_by_name(sheetname)
        elif sheetname in self._vn and self._vn[sheetname] in self._xl.sheet_names():
            sheetname = self._vn[sheetname]
            sh = self._xl.sheet_by_name(sheetname)
        else:
            return None, None
        try:
            headers = [k.value for k in sh.row(0)]
        except IndexError:
            self._print('Empty sheet %s' % sheetname)
            return sh, None
        self._print('\nOpened sheet %s' % sheetname)
        return sh, headers

    def _process_sheet(self, etype):

        sh, headers = self._sheet_accessor(etype)
        if sh is None:
            return

        for row in range(1, sh.nrows):
            rowdata = {headers[i]: self._grab_value(k) for i, k in enumerate(sh.row(row))}
            ent = self.ar[rowdata['external_ref']]
            if etype == 'quantity' and ent is None:
                try:
                    ent = self.qi.get_canonical(rowdata['external_ref'])
                    self.ar.add_entity_and_children(ent)  # this ignores EntityExists
                except EntityNotFound:
                    ent = None
            if ent is None:  # [still]
                self._new_entity(etype, rowdata)
            else:
                try:
                    rowdata.pop('external_ref')
                except KeyError:
                    continue
                rowdata.pop('uuid', None)
                ref = rowdata.pop(ent.reference_field, None)
                if self._merge == 'overwrite' and ref is not None:
                    if isinstance(ref, str) and etype == 'flow':
                        ref = self.qi.get_canonical(ref)
                    if ref is not None and ent.reference_entity != ref:
                        self._print('Updating reference entity %s -> %s' % (ent.reference_entity, ref))
                        ent[ent.reference_field] = ref
                for k, v in rowdata.items():
                    if v is None:
                        continue
                    if ent.has_property(k):
                        if self._merge == 'defer':
                            continue
                        elif ent[k] == v:
                            continue
                    self._print('Updating %s[%s] -> %s' % (ent.external_ref, k, v))
                    ent[k] = v

    def _print(self, *args):
        if self._quiet:
            return
        print(*args)

    def __enter__(self):
        """Return self object to use with "with" statement."""
        return self

    def __exit__(self, *args):
        if hasattr(self._xl, 'release_resources'):
            self._xl.release_resources()
        self._xl = None


class XlsxArchiveUpdater(XlsxUpdater):
    def __init__(self, archive, xlrd_like, merge='defer', quiet=True):
        """
        Updates entity meta-information from a spreadsheet created from above. external_ref is used as a key;
        if an entity is not found, a new one is created.  'uuid' and reference fields are only used for creation
        of new entities.
        Note: uses the default sheet names of 'flow' and 'quantity'

        :param archive:
        :param xlrd_like: XLRD workbook-like object OR filename of XLS file
        :param merge: ['defer'] - do not overwrite existing properties
                      ['overwrite'] - do overwrite existing properties
        :return:
        """
        self._ar = archive
        self._qi = archive.make_interface('quantity')
        super(XlsxArchiveUpdater, self).__init__(xlrd_like, merge=merge, quiet=quiet)

    @property
    def ar(self):
        return self._ar

    @property
    def origin(self):
        return self.ar.ref

    @property
    def qi(self):
        return self._qi

    def _new_entity(self, etype, rowdata):
        rowdata['externalId'] = rowdata.pop('external_ref')
        rowdata['entityId'] = rowdata.pop('uuid', None)
        rowdata['entityType'] = etype
        self._ar.entity_from_json(rowdata)

    def apply(self):
        for etype in ('quantity', 'flow'):  # these are the only types that are currently handled
            self._process_sheet(etype)
        self._process_flow_properties()

    def get_flow(self, flow):
        return self.ar.get(flow)

    def get_context(self, cx):
        return self.ar.tm[cx]



import json
import os
import re

from collections import defaultdict

# from antelope import MultipleReferences

from ..exchanges import AmbiguousReferenceError

from ..entities import *
from ..entities.processes import NoExchangeFound
from ..archives import LcArchive
from .file_store import FileStore
from .parse_math import parse_math


geog_tail = re.compile('\\b[A-Z]+[o-]?[A-Z]*$')  # capture, e.g. 'ZA', 'GLO', 'RoW', 'US-CA' but not 'PET-g'

def pull_geog(flowname):
    raise NotImplementedError
    '''
    try:
        return geog_tail.search(flowname).group()
    except AttributeError:
        return None
    '''


valid_types = {'processes', 'flows', 'flow_properties'}


class OpenLcaException(Exception):
    pass


class _NotAnRx(Exception):
    pass


SKIP_DURING_INDEX = ('context.json', 'meta.info')


class OpenLcaJsonLdArchive(LcArchive):
    """
    Opens JSON-LD archives formatted according to the OpenLCA schema
    """
    def _cat_as_list(self, cat_id):
        cat = self._cat_index[cat_id]
        if 'category' in cat:
            return self._cat_as_list(cat['category']['@id']) + [cat['name']]
        return [cat['name']]

    def _gen_index(self):
        self._print('Generating index')
        self._type_index = dict()
        self._lm_index = dict()
        # these are cheap- but we should really not need more than 3
        self._cat_index = dict()  # maps olca id to olca category object
        self._cat_lookup = dict()  # maps tuple of category-as-list to olca id
        self._cat_tuple = dict()  # maps olca id to tuple of category-as-list

        for f in self._archive.listfiles():
            if f in SKIP_DURING_INDEX:
                continue
            ff = f.split('/')
            if len(ff) < 2:
                self._type_index['root'] = ff[0]
                continue
            fg = ff[1].split('.')
            self._type_index[fg[0]] = ff[0]
            if ff[0] == 'lcia_methods':
                # we need to do a further indexing to generate a list of LCIA quantities aka categories by method.
                # that means actually loading the files
                obj = self._create_object(ff[0], fg[0])
                for c in obj['impactCategories']:
                    self._lm_index[c['@id']] = fg[0]
            elif ff[0] == 'categories':
                obj = self._create_object(ff[0], fg[0])
                self._cat_index[fg[0]] = obj
        for cat_key in self._cat_index.keys():
            cat = self._cat_as_list(cat_key)
            lookup_key = tuple(cat)
            self._cat_tuple[cat_key] = lookup_key  # forward lookup key -> tuple
            self._cat_lookup[lookup_key] = cat_key  # reverse lookup of tuple -> key
            self.tm.add_context(lookup_key, cat_key)

    def __init__(self, source, prefix=None, skip_index=False, **kwargs):
        super(OpenLcaJsonLdArchive, self).__init__(source, **kwargs)

        self._drop_fields['process'].extend(['processDocumentation'])

        self._archive = FileStore(source, internal_prefix=prefix)

        self._type_index = None
        if not skip_index:
            self._gen_index()

    def _check_id(self, _id):
        return self[_id] is not None

    def _create_object(self, typ, key):
        return json.loads(self._archive.readfile(os.path.join(typ, key + '.json')))

    def _process_from_json(self, entity_j, uid):
        process = super(OpenLcaJsonLdArchive, self)._process_from_json(entity_j, uid)
        if process.has_property('allocationFactors'):
            # we do not need to replicate 0-valued allocation factors, esp. when there are thousands of them
            process['allocationFactors'] = [k for k in process['allocationFactors'] if k['value'] != 0]
        return process

    def _clean_object(self, typ, key):
        """

        :param typ:
        :param key:
        :return: clean json, name, list of category IDs
        """
        j = self._create_object(typ, key)
        j.pop('@context')
        j.pop('@id')
        name = j.pop('name')

        if 'category' in j:
            c_j = j.pop('category')
            cat = self._get_category_list(c_j['@id'])
        else:
            cat = []
        return j, name, cat

    def _get_category_list(self, cat_key):
        if cat_key in self._cat_tuple:
            return list(self._cat_tuple[cat_key])
        c_j = self._cat_index[cat_key]
        if 'category' in c_j:
            cat = self._get_category_list(c_j['category']['@id'])
        else:
            cat = []
        cat.append(c_j['name'])
        return cat

    def openlca_category(self, name_or_full_list):
        """
        returns an olca id for a name or a list of hierarchical names, if one exists
        :param name_or_full_list:
        :return: either a known context or raise KeyError
        """
        try:
            return self._cat_lookup[tuple(name_or_full_list)]
        except KeyError:
            cx = self.tm[name_or_full_list]
            if cx is None:
                raise KeyError
            return self._cat_lookup[tuple(cx.as_list())]

    @property
    def openlca_categories(self):
        for k in sorted(self._cat_lookup.keys()):
            yield k

    def _create_unit(self, unit_id):
        try:
            u_j = self._create_object('unit_groups', unit_id)
        except FileNotFoundError:
            return LcUnit(unit_id), None
        unitconv = dict()
        unit = None

        for conv in u_j['units']:
            is_ref = conv.pop('referenceUnit', False)
            name = conv.pop('name')
            cf_i = conv.pop('conversionFactor')
            unitconv[name] = 1.0 / cf_i

            if is_ref:
                assert cf_i == 1, 'non-unit reference unit found! %s' % unit_id
                unit = LcUnit(name)

        if unit is None:
            raise OpenLcaException('No reference unit found for id %s' % unit_id)

        return unit, unitconv

    def _create_quantity(self, q_id):
        q = self[q_id]
        if q is not None:
            return q

        q_j, name, cat = self._clean_object('flow_properties', q_id)
        ug = q_j.pop('unitGroup')
        unit, unitconv = self._create_unit(ug['@id'])

        q = LcQuantity(q_id, Name=name, ReferenceUnit=unit, UnitConversion=unitconv, Category=cat, **q_j)

        self.add(q)
        return q

    def _create_allocation_quantity(self, process, alloc_type):
        key = '%s_%s' % (process.name, alloc_type)
        name = '%s (%s)' % (alloc_type, process.name.strip())
        u = self._ref_to_nsuuid(key)
        q = self[u]
        if q is not None:
            return q

        unit, _ = self._create_unit('alloc')
        q = LcQuantity(key, Name=name, ReferenceUnit=unit)
        self.add(q)
        assert q.uuid == u
        return q

    def _create_flow(self, f_id):
        q = self[f_id]
        if q is not None:
            return q

        f_j, name, comp = self._clean_object('flows', f_id)
        cas = f_j.pop('cas', '')
        loc = f_j.pop('location', {'name': 'GLO'})['name']

        fps = f_j.pop('flowProperties')

        qs = []
        facs = []
        ref_q = None

        for fp in fps:
            q = self.retrieve_or_fetch_entity(fp['flowProperty']['@id'], typ='flow_properties')
            ref = fp.pop('referenceFlowProperty', False)
            fac = fp.pop('conversionFactor')
            if ref:
                assert fac == 1.0, 'Non-unit reference flow property found! %s' % f_id
                ref_q = q
            else:
                if q not in qs:
                    qs.append(q)
                    facs.append(fac)
        if ref_q is None:
            raise OpenLcaException('No reference flow property found: %s' % f_id)
        f = LcFlow(f_id, Name=name, Compartment=comp, CasNumber=cas, ReferenceQuantity=ref_q, **f_j)  # context gets set by _catch_context()

        self.add(f)  # context gets matched inside tm.add_flow().  NONSPECIFIC entries are automatically prepended with parent name in CompartmentManager.new_entry()

        for i, q in enumerate(qs):
            self.tm.add_characterization(f.link, ref_q, q, facs[i], context=f.context, location=loc)

        return f

    def _add_exchange(self, p, ex):
        flow = self.retrieve_or_fetch_entity(ex['flow']['@id'], typ='flows')
        value = ex['amount']
        dirn = 'Input' if ex['input'] else 'Output'

        fp = self.retrieve_or_fetch_entity(ex['flowProperty']['@id'], typ='flow_properties')

        try:
            v_unit = ex['unit']['name']
        except KeyError:
            print('%s: %d No unit! using default %s' % (p.external_ref, ex['internalId'], fp.unit))
            v_unit = fp.unit

        if v_unit != fp.unit:
            oldval = value
            value *= fp.convert(from_unit=v_unit)
            self._print('%s: Unit Conversion exch: %g %s to native: %g %s' % (p.uuid, oldval, v_unit, value, fp.unit))

        if fp != flow.reference_entity:
            try:
                value /= fp.cf(flow)  ## is this even right?  ### yes  # TODO: account for locale?
            except (TypeError, ZeroDivisionError):
                print('%s:%s:%s flow reference quantity does not match\n%s exchange f.p. Conversion Required' %
                      (p.external_ref, dirn, flow.external_ref, flow.name))
                print('From %s to %g %s' % (flow.unit, value, fp.unit))
                val = parse_math(input('Enter conversion factor 1 %s = x %s [context %s]\nx: ' %
                                       (flow.unit, fp.unit, flow.context)))
                self.tm.add_characterization(flow.link, flow.reference_entity, fp, val, context=flow.context,
                                             origin=self.ref)
                value /= fp.cf(flow)

        is_ref = ex.pop('quantitativeReference', False)
        if is_ref:
            term = None
        else:
            cx = self.tm[flow.context]
            if cx.elementary:
                term = cx
            else:
                if 'defaultProvider' in ex:
                    term = ex['defaultProvider']['@id']
                else:
                    term = cx

        exch = p.add_exchange(flow, dirn, value=value, termination=term, add_dups=True)
        if is_ref:
            p.set_reference(flow, dirn)

        if 'description' in ex:
            exch.comment = ex['description']

        return exch

    def _get_rx(self, p, flow_ref):
        rf = self.retrieve_or_fetch_entity(flow_ref)
        try:
            ft = rf['flowType']
            if ft == 'ELEMENTARY_FLOW':
                print('%s: Skipping allocation factor for elementary flow %s' % (p.external_ref, rf.external_ref))
                raise _NotAnRx
        except KeyError:
            ft = 'PRODUCT_FLOW'  # more common ??
        dr = {'PRODUCT_FLOW': 'Output',
              'WASTE_FLOW': 'Input'}[ft]
        try:
            rx = p.reference(rf)
        except NoExchangeFound:
            # implicit trickery with schema: reference flows MUST be outputs for products, inputs for wastes
            rx_cands = list(_x for _x in p.exchange_values(rf, direction=dr) if _x.type in ('context', 'cutoff'))
            if len(rx_cands) == 0:
                print('%s: Unable to find allocatable exchange for %s' % (p.external_ref, rf.external_ref))
                raise _NotAnRx
            elif len(rx_cands) > 1:
                raise AmbiguousReferenceError('%s: Multiple flows with ID %s' % (p.external_ref, rf.external_ref))
            else:
                rx = rx_cands[0]
            p.set_reference(rx.flow, rx.direction)

        return rx

    def _apply_olca_allocation(self, p, alloc=None):
        """
        For each allocation factor, we want to characterize the flow so that its exchange value times its
        characterization equals the stated factor.  Then we want to allocate the process by its default allocation
        property.

        A few notes: the allocation quantity is PROCESS-SPECIFIC (strictly, it's process.name-specific, so that is
        either a feature or a bug, TBD).  Because of this, NO allocation will be achievable UNLESS the process has
        alloationFactors defined.

        This whole thing needs to be sorted out with testing, probably in collaboration with GreenDelta.
        :param p: an LcProcess generated from the JSON-LD archive
        :return:
        """
        if alloc is None:
            return
        if p.has_property('defaultAllocationMethod'):
            dm = p['defaultAllocationMethod']
        else:
            dm = 'NO_ALLOCATION'
        _causal_msg = True
        stored_alloc = []
        for af in alloc:
            try:
                rx = self._get_rx(p, af['product']['@id'])
            except _NotAnRx:
                continue
            if af['allocationType'] == 'CAUSAL_ALLOCATION':
                if af['value'] == 0:
                    # Keep 0-allocation factors for non-causal
                    continue

                if dm != 'CAUSAL_ALLOCATION':
                    if _causal_msg:
                        print('%s: Skipping Speculative CAUSAL_ALLOCATION' % p.external_ref)
                        _causal_msg = False
                    continue
                f = self.retrieve_or_fetch_entity(af['exchange']['flow']['@id'])

                xs = list(p.exchange_values(f))
                if len(xs) > 1:
                    raise AmbiguousReferenceError('%s: Multiple flows with ID %s' % (p.external_ref, f.external_ref))
                x = xs[0]

                val = af['value']
                stored_alloc.append(af)

                x[rx] = x.value * val
                if _causal_msg:
                    print('%s: Warning: causal allocation has not been tested' % p.external_ref)
                    _causal_msg = False
            else:
                q = self._create_allocation_quantity(p, af['allocationType'])

                v = af['value'] / rx.value
                stored_alloc.append(af)

                if v != 0:
                    if not rx.is_reference:
                        p.set_reference(rx.flow, rx.direction)

                self.tm.add_characterization(rx.flow.link, rx.flow.reference_entity, q, v,
                                             context=rx.flow.context, origin=self.ref)
                #f.add_characterization(q, value=v)

        if dm != 'NO_ALLOCATION':
            aq = self._create_allocation_quantity(p, p['defaultAllocationMethod'])
            p.allocate_by_quantity(aq)
        p['allocationFactors'] = stored_alloc  # only keep factors we used

    def _create_process(self, p_id):
        q = self[p_id]
        if q is not None:
            return q

        p_j, name, cls = self._clean_object('processes', p_id)
        ss = p_j.pop('location', {'name': 'GLO'})['name']
        stt = dict()
        for key, tgt in (('validFrom', 'begin'), ('validUntil', 'end')):
            try:
                stt[tgt] = p_j['processDocumentation'][key]
            except KeyError:
                pass

        exch = p_j.pop('exchanges')

        if 'allocationFactors' in p_j:
            # process later-- do we keep or remove them???
            alloc = p_j.pop('allocationFactors')
        else:
            alloc = None


        p = LcProcess(p_id, Name=name, Classifications=cls, SpatialScope=ss, TemporalScope=stt, **p_j)

        self.add(p)

        for ex in exch:
            self._add_exchange(p, ex)

        self._apply_olca_allocation(p, alloc)

        return p

    def _create_lcia_quantity(self, l_j, method, **kwargs):
        q_id = l_j['@id']
        l = self[q_id]
        if l is not None:
            return l

        l_obj, l_name, cats = self._clean_object('lcia_categories', q_id)
        c_desc = l_obj.pop('description', None)
        ver = l_obj.pop('version', None)
        indicator = l_obj.pop('referenceUnitName')
        unit = LcUnit(indicator)

        q_name = ', '.join([method, l_name])

        q = LcQuantity(q_id, Name=q_name, ReferenceUnit=unit, Method=method, Category=l_name, Indicator=indicator,
                       CategoryDescription=c_desc, Version=ver, **kwargs)

        self.add(q)
        for factor in l_obj.get('impactFactors', []):
            flow = self._create_flow(factor['flow']['@id'])
            loc = factor.get('location')
            ''' # this does not work
            if loc is None:
                try:
                    loc = geog_tail.search(flow.name).group()
                except AttributeError:
                    pass
            '''

            ref_qty = self._create_quantity(factor['flowProperty']['@id'])
            assert flow.reference_entity == ref_qty
            # value = factor['value']

            self.tm.add_characterization(flow.link, ref_qty, q, factor['value'], context=flow.context, location=loc,
                                         origin=self.ref)

        return q

    def _create_lcia_category(self, c_id):
        """
        In this case, the client has requested a specific lcia quantity, a.k.a. LCIA category.  This is tricky because
        we don't know which method the category is part of. So we modify our indexing process to track this.
        :param c_id:
        :return:
        """
        if c_id in self._lm_index:
            try:
                q = next(t for t in self._create_lcia_method(self._lm_index[c_id], c_id) if t.uuid == c_id)
            except StopIteration:
                raise OpenLcaException('Specified LCIA category does not match the one found')
            return q
        raise KeyError('Specified key is not an LCIA Category: %s' % c_id)

    def _create_lcia_method(self, m_id, *categories):
        """
        Note: in OLCA archives, an "LCIA Method" is really a methodology with a collection of category indicators, which
        is what we colloquially call "methods". So every method includes zero or more distinct quantities.
        :param m_id:
        :return:
        """
        m_obj, method, cats = self._clean_object('lcia_methods', m_id)
        m_desc = m_obj.pop('description', None)

        sets = []
        norms = defaultdict(list)
        weights = defaultdict(list)

        if 'nwSets' in m_obj:
            for n in m_obj['nwSets']:
                norm_j = self._create_object('nw_sets', n['@id'])
                sets.append(norm_j['name'])
                for fac in norm_j['factors']:
                    norms[fac['impactCategory']['@id']].append(fac.get('normalisationFactor', None))
                    weights[fac['impactCategory']['@id']].append(fac.get('weightingFactor', None))

        qs = []

        for imp in m_obj.pop('impactCategories', []):
            if len(categories) > 0:
                # if the user specifies certain categories to load, then skip the ones that aren't specified
                if imp['@id'] not in categories:
                    continue
            q = self._create_lcia_quantity(imp, method, MethodDescription=m_desc)
            norm = norms[q.external_ref]
            if len(norm) > 0:
                q['normalisationFactors'] = norm
                q['normSets'] = sets
                q['weightingFactors'] = weights[q.external_ref]
            qs.append(q)

        return qs

    def _fetch(self, key, typ=None, **kwargs):
        if typ is None:
            if self._type_index is None:
                self._gen_index()
            typ = self._type_index[key]
        try:
            _ent_g = {'processes': self._create_process,
                      'flows': self._create_flow,
                      'flow_properties': self._create_quantity,
                      'lcia_methods': self._create_lcia_method,
                      'lcia_categories': self._create_lcia_category}[typ]
        except KeyError:
            print('Warning: generating generic object for unrecognized type %s' % typ)
            _ent_g = lambda x: self._create_object(typ, x)

        return _ent_g(key)

    def _load_all(self, **kwargs):
        self._print('Loading processes')
        for f in self._archive.listfiles(in_prefix='processes'):
            ff = f.split('/')
            fg = ff[1].split('.')
            self._create_process(fg[0])
        self._print('Loading LCIA methods')
        for f in self._archive.listfiles(in_prefix='lcia_methods'):
            ff = f.split('/')
            fg = ff[1].split('.')
            self._create_lcia_method(fg[0])



"""
Initialization file for antelope_core package

Package Includes:
 - Exchanges, Characterizations, and Lcia Results
 - LCA Process, Flow, and Quantity entities
 - An Entity Store (which should really be a database instead of a python object) and subclasses for life cycle data
"""


from synonym_dict import LowerDict
from antelope import antelope_herd

import importlib

from .from_json import from_json, to_json
from .archives import archive_factory, ArchiveError

FOUND_PROVIDERS = LowerDict()
def _find_providers():
    for ant in [__name__] + antelope_herd:
        try:
            importlib.import_module('.', package=ant)
        except ModuleNotFoundError:
            continue
        p = importlib.import_module('.providers', package=ant)
        try:
            provs = getattr(p, 'PROVIDERS')
        except AttributeError:
            print('No PROVIDERS found in %s' % ant)
            continue
        for ds_type in provs:
            FOUND_PROVIDERS[ds_type] = p

    print('Found Antelope providers:' )
    for k, v in FOUND_PROVIDERS.items():
        print('%s:%s' % (v.__name__, k))


def herd_factory(ds_type):
    try:
        return archive_factory(ds_type)
    except ArchiveError:
        if len(FOUND_PROVIDERS) == 0:
            _find_providers()
        if ds_type in FOUND_PROVIDERS:
            prov = FOUND_PROVIDERS[ds_type]
            try:
                return getattr(prov, ds_type)
            except AttributeError:
                dsl = ds_type.lower()
                try:
                    attr = next(k for k in prov.PROVIDERS if k.lower().startswith(dsl))
                    return getattr(prov, attr)
                except (StopIteration, AttributeError):
                    raise ArchiveError('ds_type %s not found in %s' % (ds_type, prov.__name__))
    print('# LENGTH OF PROVIDERS: %d' % len(FOUND_PROVIDERS))
    raise ImportError('Cannot find a package for loading %s' % ds_type)


def add_antelope_providers(mod, provs=None):
    """
    Manually register Antelope providers
    :param mod: a python module or package that contains data providers
    :param provs: a list of provider class names that can be imported from module (if omitted,
    attempts to getattr(module, 'PROVIDERS')
    (note: may also supply a single class as a string- it will get listified)
    :return:
    """
    if isinstance(provs, str):
        provs = [provs]
    if len(FOUND_PROVIDERS) == 0:
        _find_providers()
    if provs is None:
        try:
            provs = getattr(mod, 'PROVIDERS')
        except AttributeError:
            print('No providers found in %s' % mod.__name__)
            provs = []
    for ds_type in provs:
        print('Adding %s:%s' % (mod.__name__, ds_type))
        FOUND_PROVIDERS[ds_type] = mod


from .catalog import LcCatalog
from .lc_resource import LcResource
from .data_sources.local import make_config



"""
Archive testing is divided up loosely into pre-entity (e.g. test_entity_store and test_base) and post-entity testing
(e.g. this file).  Of course test_base has entities but they are more for the purpose of validating name setting.

This class imports a minimal archive file created in antelope_utilities to contain a multioutput, nontrivial process
(USLCI petroleum refining) and an intermediate process (a grid mix).  These don't do anything special but the refinery
process can test both exchange generation / allocation and lcia.

Subclass this to access that archive.  Current uses:
"""

import unittest
import os

from ...archives import LcArchive


refinery_archive = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_archive.json')
grid_id = '96bffbb9-b875-36cf-8a11-5723c9d239d9'
petro_id = '0aaf1e13-5d80-37f9-b7bb-81a6b8965c71'


class BasicEntityTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.A = LcArchive.from_file(refinery_archive)
        cls.grid = cls.A[grid_id]
        cls.petro = cls.A[petro_id]



from .basic import BasicImplementation
from .background import BackgroundImplementation, search_skip
from .configure import CoreConfigureImplementation, ConfigureImplementation, LcConfigureImplementation
from .quantity import QuantityImplementation
from .index import IndexImplementation
from .exchange import ExchangeImplementation



"""
Ecoinvent LCI testing!

In order to run this test, you need to have a catalog that can resolve ecoinvent references.  Configure this by
setting a value for the parameter CATALOG_ROOT below, just after the imports.  If this is not defined, the tests will
be skipped.

"""

import os
import unittest
import random

from collections import namedtuple

from math import isclose

from antelope import EntityNotFound


from ...entities import LcProcess
from ...archives import LcArchive, archive_from_json

from ..local import check_enabled


EcoinventNode = namedtuple('EcoinventNode', ['version', 'model', 'node'])
_debug = False


if __name__ == '__main__':
    _run_ecoinvent = check_enabled('ecoinvent')
else:
    _run_ecoinvent = check_enabled('ecoinvent') or _debug


if _run_ecoinvent:
    from .test_aa_local import cat
else:
    cat = None


def lci_cache_file(version, model):
    return os.path.join(cat.archive_dir, 'lci_cache_%s_%s.json.gz' % (version, model))


def inv_cache_file(version, model):
    """
    :param version:
    :param model:
    :return:
    """
    return os.path.join(cat.archive_dir, 'inv_cache_%s_%s.json.gz' % (version, model))


def test_ref(version, model):
    return 'test.lci.ecoinvent.%s.%s' % (version, model)


def find_lci_ref(version, model):
    check_ref = 'ecoinvent.lci.%s.%s' % (version, model)
    if check_ref in cat.references:
        return check_ref
    check_ref = '.'.join(['local', check_ref])
    if check_ref in cat.references:
        return check_ref
    return None


def _extract_and_reduce_lci(node):
    """
    This function performs the (surpassingly slow) task of extracting reduced aggregated LCI results from the large
    bundled ecoinvent 7z LCI archives.  (The operation is slow on RAM-limited machines because the 7z algorithm requires
    tremendous memory)

    The routine requires the process inventory test case to be selected manually.

    The selected inventory is loaded, and then one hundred exchanges are selected at random and the rest are removed.
    This reduces the file size (and load time) of the generated archives without sacrificing the representativeness of
    the computation.

    :param node:
    :return:
    """
    filename = lci_cache_file(node.version, node.model)
    ref = test_ref(node.version, node.model)
    if os.path.exists(filename):
        a = archive_from_json(filename, ref=ref)
    else:
        a = LcArchive(filename, ref=ref)

    if ref not in cat.references:
        cat.add_existing_archive(a, interfaces='inventory', static=True)

    try:
        cat.query(ref).get(node.node)
        return
    except EntityNotFound:
        pass

    cat.get_resource(ref, 'inventory').remove_archive()

    lci_ref = find_lci_ref(node.version, node.model)
    if lci_ref is None:
        print('No LCI resource for (%s, %s)' % (node.version, node.model))
        return

    print('WARNING: extracting and reducing LCI data can be very slow >60s per file')

    p_ref = cat.query(lci_ref).get(node.node)

    if p_ref is None:
        print('No process found with reference %s' % node.node)
        return

    p_rx = next(p_ref.references())
    exchs = random.sample([_x for _x in p_ref.inventory(ref_flow=p_rx)], 100)

    p_slim = LcProcess(p_ref.uuid, Name=p_ref['Name'])
    p_slim.add_exchange(p_rx.flow, p_rx.direction, value=1.0)  # value=p_ref.reference_value(p_rx.flow)) ##
    p_slim.set_reference(p_rx.flow, p_rx.direction)

    for x in exchs:
        p_slim.add_exchange(x.flow, x.direction, value=x.value, termination=x.termination)

    a.add_entity_and_children(p_slim)
    a.write_to_file(filename, complete=True, gzip=True)


class EcoinventLciTest(unittest.TestCase):
    _nodes = {
        EcoinventNode('3.2', 'apos', '18085d22-72d0-4588-9c69-7dbeb24f8e2f'),
        # EcoinventNode('3.2', 'apos', 'ca4a6d8a-2399-4645-ac20-17343c694f2b'),  # potato seed, for setting- fg scc member
        EcoinventNode('3.2', 'conseq', '6b0f32fe-329d-4c1f-9205-0ea78f4f42e5')
    }

    @classmethod
    def nodes(cls):
        for node in cls._nodes:
            if node.node is None:
                continue
            yield node

    @classmethod
    def setUpClass(cls):
        if _run_ecoinvent:
            for node in cls.nodes():
                if node.node is None:
                    continue
                ref = test_ref(node.version, node.model)
                if ref in cat.references:
                    try:
                        cat.query(ref).get(node.node)
                    except EntityNotFound:
                        _extract_and_reduce_lci(node)
                    continue
                # if ref is not in cat, or if node is not present, then we need to populate node
                _extract_and_reduce_lci(node)

    def setUp(self):
        if not _run_ecoinvent:
            self.skipTest('Ecoinvent not setup')

    def test_lci(self):
        print('Run Test? %s' % _run_ecoinvent)
        for node in self.nodes():
            lci_result = cat.query(test_ref(node.version, node.model)).get(node.node)
            rx = lci_result.reference()
            challenge = cat.query('local.ecoinvent.%s.%s' % (node.version, node.model), debug=False).get(node.node)

            c_lci = challenge.lci(ref_flow=rx.flow.external_ref, threshold=1e-10)
            lci_check = {x.flow.external_ref: x for x in c_lci}

            count = 0
            inverted = []
            fail = []

            for i in lci_result.inventory(rx):  # inventory method normalizes in ExchangeValue.from_allocated
                count += 1
                z = lci_check[i.flow.external_ref]
                if isclose(i.value, z.value, rel_tol=1e-6):
                    continue
                elif isclose(i.value, -z.value, rel_tol=1e-6):
                    inverted.append(i)
                else:
                    fail.append(i)
                    print('Not close! %s\nInv: %12g Chk: %12g ratio: %.10f\n' % (i.flow, i.value, z.value,
                                                                                 i.value / z.value))
            self.assertTrue(len(fail) < 2)
            self.assertEqual(len(inverted), 0)

            print('Total %d, inverted %d, fail %d' % (count, len(inverted), len(fail)))

"""
        for v, m in ec.valid_lci():
            if not os.path.exists(lci_cache_file(v, m)):
                continue
            lci_a = archive_from_json(lci_cache_file(v, m))
            p = next(lci_a.entities_by_type('process'))

            q = cat.query(ref_(v, m, 'inventory')).lci(p.external_ref)



        true_dirn = [_x for _x in bm.foreground(my_p)][0].direction
        if my_p.reference().direction != true_dirn:
            mult = -1
        else:
            mult = 1

        fail = []
        for i in lci:
            try:
                ic = inv_check_d[i.key]
            except KeyError:
                continue
            if not isclose(i.value, mult*ic.value, rel_tol=1e-7):
                print('Not close! %s\nInv: %12g Chk: %12g ratio: %.10f\n' % (i.flow, i.value, ic.value, 
                                                                             i.value / ic.value))
                fail.append(i)
        
        print('Failed: %d' % len(fail))


"""

if __name__ == '__main__':
    unittest.main()



from .data_source import DataSource, ResourceInfo
# from antelope_utilities import IPCC_2007_TRACI_REF
from ..lcia_engine import IPCC_2007_GWP


IPCC_2007_TRACI_REF = 'lcia.ipcc.2007.traci21'


gwp_ipcc_2007 = ResourceInfo(IPCC_2007_GWP, 'json', None, None,
                             {'hints': [['context', 'air', 'to air']]}, {})


class GwpIpcc2007(DataSource):
    _ds_type = 'json'

    @property
    def references(self):
        """
        There's a precedence issue here, becaue the class MUST yield the same reference name as what's encoded in
        the JSON file, but it is wasteful to load the JSON file just to learn the canonical ref.  The "proper" way to
        do this is to [better] manage the workflow by which the JSON file is created (i.e. in antelope_utilities)
        and then use the same ref for both paths.

        heh- I hint at this challenge in the definite source by naming it AUTHORIZED_REF
        :return:
        """
        yield IPCC_2007_TRACI_REF

    def interfaces(self, ref=IPCC_2007_TRACI_REF):
        for i in ('index', 'quantity'):
            yield i

    def make_resources(self, ref=IPCC_2007_TRACI_REF):
        if ref in self.references:
            yield self._make_resource(ref, info=gwp_ipcc_2007, interfaces=self.interfaces(), static=True)



import unittest

from ...entities import LcFlow
from ..ecospold import EcospoldV1Archive, apply_conversion


class EcospoldV1Test(unittest.TestCase):
    def test_unit_conversion(self):
        ar = EcospoldV1Archive('/tmp')
        kbq = ar._create_quantity('kBq')
        bq = ar._create_quantity('Bq')
        f = LcFlow(1, referenceQuantity=bq, Name='my bq flow')
        ar.add(f)
        self.assertEqual(kbq.cf(f), 0.0)
        self.assertTrue(apply_conversion(kbq, f))
        self.assertEqual(kbq.cf(f), 0.001)


if __name__ == '__main__':
    unittest.main()



import os
import re
from lxml import objectify

from .data_source import DataSource, DataCollection
from .ecoinvent_lcia import EcoinventLciaConfig, EI_LCIA_SPREADSHEETS
from ..providers.file_store import FileStore

FILE_PREFIX = ('current_Version_', 'ecoinvent ')
FILE_EXT = ('7z', 'zip')
ECOINVENT_SYS_MODELS = ('apos', 'conseq', 'cutoff', 'undefined')
MODELMAP = {
    'apos': ('apos',),
    'conseq': ('consequential', 'consequential_longterm'),
    'cutoff': ('cutoff', 'cut-off'),
    'undefined': ('undefined', )
}

E_CFG = {'hints': [  # cover elementary contexts that need directional hints, plus units used in elementary flows only
    ['context', 'air', 'to air'],
    ['context', 'water', 'to water'],
    ['quantity', 'EcoSpold Quantity kg', 'Mass'],
    ['quantity', 'EcoSpold Quantity m2', 'Area'],
    ['quantity', 'EcoSpold Quantity MJ', 'Net Calorific Value'],
    ['quantity', 'EcoSpold Quantity m2*year', 'Area*time'],
    ['quantity', 'EcoSpold Quantity m3', 'Volume']
]}  # omitted: kBq, EUR2005, m3*year


class Ecoinvent2Base(DataSource):

    _ds_type = 'EcospoldV1Archive'

    def __init__(self, data_root, version, **kwargs):
        super(Ecoinvent2Base, self).__init__(data_root=data_root, **kwargs)
        self._version = version

    @property
    def references(self):
        """
        Generates a list of semantic references the DataSource knows how to instantiate
        :return:
        """
        yield 'local.ecoinvent.%s' % self._version

    def interfaces(self, ref):
        """
        Generates a list of interfaces known for the given reference. the reference must be in the list of references.
        :param ref:
        :return:
        """
        for k in ('exchange', 'quantity'):
            yield k

    @property
    def source(self):
        return os.path.join(self.root, self._version, 'datasets')

    def make_resources(self, ref):
        """
        Generates an exhaustive sequence of LcResource objects for a given reference.
        :param ref:
        :return:
        """
        ref = next(self.references)
        yield self._make_resource(ref, self.source, interfaces=tuple(self.interfaces(ref)), config=E_CFG)


class Ecoinvent3Base(DataSource):

    _ds_type = 'EcospoldV2Archive'

    def __init__(self, data_root, version, model, **kwargs):
        assert model in ECOINVENT_SYS_MODELS
        super(Ecoinvent3Base, self).__init__(data_root=data_root, **kwargs)
        self._version = version
        self._model = model

    @property
    def _lci_ref(self):
        if self.lci_source is not None:
            yield 'local.ecoinvent.lci.%s.%s' % (self._version, self._model)

    @property
    def _inv_ref(self):
        if self.inv_source is not None:
            yield 'local.ecoinvent.%s.%s' % (self._version, self._model)

    @property
    def references(self):
        for x in self._lci_ref:
            yield x
        for x in self._inv_ref:
            yield x

    def interfaces(self, ref):
        yield 'exchange'

    def make_resources(self, ref):
        if ref in self._lci_ref:
            yield self._make_resource(ref, self.lci_source, interfaces='exchange', prefix='datasets', config=E_CFG)
        elif ref in self._inv_ref:
            if self._model == 'undefined':
                yield self._make_resource(ref, self.inv_source, interfaces='exchange', prefix='datasets - public',
                                          linked=False, config=E_CFG)
            else:
                yield self._make_resource(ref, self.inv_source, interfaces='exchange', prefix='datasets', config=E_CFG)

    def _fname(self, ftype=None):
        precheck = os.path.join(self.root, self._model)
        if ftype is not None:
            precheck += '_%s' % ftype
        if os.path.isdir(precheck):
            return precheck
        for pf in FILE_PREFIX:
            for mod in MODELMAP[self._model]:
                if ftype is None:
                    pname = '%s%s_%s_ecoSpold02' % (pf, self._version, mod)
                else:
                    pname = '%s%s_%s_%s_ecoSpold02' % (pf, self._version, mod, ftype)
                dsource = os.path.join(self.root, self._version, pname)
                if os.path.isdir(dsource):
                    return dsource
                for ext in FILE_EXT:
                    fname = '%s.%s' % (pname, ext)
                    source = os.path.join(self.root, self._version, fname)
                    if os.path.exists(source):
                        return source

    @property
    def inv_source(self):
        return self._fname()

    @property
    def lci_source(self):
        return self._fname('lci')

    def elementary_exchanges(self):
        for k in (self.inv_source, self.lci_source):
            try:
                fs = FileStore(k, internal_prefix='MasterData')
            except TypeError:
                continue
            try:
                o = objectify.fromstring(fs.readfile('ElementaryExchanges.xml'))
            except FileNotFoundError:
                continue
            yield o


class EcoinventConfig(DataCollection):
    """
    Ecoinvent Configurator
    This DataCollection generates LcResource objects for ecoinvent archives.  The only required input is the root
    folder containing the ecoinvent data.  Within this folder should be subfolders named after the major and minor
    version of the database (e.g. "3.3"). Within the subfolders should be the archives.

    Archives must be named according to ecoinvent's various conventions:
     #1#2_#3_ecoSpold02#4
    where
     #1 is one of ('current_Version_' or 'ecoinvent ')
     #2 is the major and minor version ('2.2', '3.01', '3.1', '3.2', etc)
     #3 is the system model ('apos', 'cutoff', 'consequential', or 'consequential_longterm')
     #4 is either omitted (in the case of an expanded directory) or an archive extension ('.zip' or '.7z')

    Within the archives, either compressed or expanded, the datasets are assumed to be in a subfolder called 'datasets'

    The class does not currently support loading undefined data collections, but this could be added in the future.

    """
    @property
    def ecoinvent_versions(self):
        for d in os.listdir(self._root):
            if os.path.isdir(os.path.join(self._root, d)):
                if re.match('[23]\.[0-9]+', d):
                    yield d
                if d.lower() == 'lcia':
                    yield d

    def factory(self, data_root, **kwargs):
        for v in self.ecoinvent_versions:
            if v.lower() == 'lcia':
                lcia_path = os.path.join(data_root, v)
                for ver, filename in EI_LCIA_SPREADSHEETS.items():
                    if os.path.exists(os.path.join(lcia_path, filename)):
                        yield EcoinventLciaConfig(lcia_path, version=ver)
            elif v.startswith('2'):
                yield Ecoinvent2Base(data_root, v, **kwargs)
            else:
                for m in ECOINVENT_SYS_MODELS:
                    yield Ecoinvent3Base(data_root, v, m, **kwargs)
                    yield EcoinventLciaConfig(os.path.join(data_root, v), version=v)

    def elementary_exchanges(self):
        print('DEPRECATED. please use configured EcospoldV2 Archives instead. (??)')
        for k, v in sorted(self._sources.items(), key=lambda x: x[0], reverse=True):
            if hasattr(v, 'elementary_exchanges'):
                for e in v.elementary_exchanges():
                    yield e
